// Copyright (c) 2018 Kerlink SA. All rights reserved.

#include "api.h"

#include <algorithm>
#include <chrono>
#include <cstring>
#include <ctime>
#include <iterator>
#include <ostream>
#include <stdexcept>

#include "third-party/variant.h"

#include "utils/logger.h"
#include "utils/type_traits.h"

#include "protos/control-rpc.pb.h"
#include "protos/control-statistic.pb.h"
#include "protos/control-verbosity.pb.h"
#include "protos/control-version.pb.h"
#include "protos/downlink-container.pb.h"
#include "protos/uplink-container.pb.h"

namespace zmq {
namespace {
std::ostream &zendl(std::ostream &_os)
{
	_os << ": " << zmq_strerror(zmq_errno()) << std::endl;
	return _os;
}
}

Message::Message(void) { zmq_msg_init(&m_message); }

Message::~Message(void)
{
	if (zmq_msg_close(&m_message) < 0) {
		logr::warning << "Failed to close zmq message" << zendl;
	}
}

bool Message::reserve(std::size_t _size)
{
	if (zmq_msg_init_size(&m_message, _size) < 0) {
		logr::error << "Failed to initialize zmq message" << zendl;
		return false;
	}

	return true;
}

bool Message::assign(const void *_data, std::size_t _size)
{
	if (!reserve(_size)) {
		return false;
	}

	std::memcpy(data(), _data, _size);

	return true;
}

bool Message::assign(const std::string &_message)
{
	return assign(_message.data(), _message.size());
}

void *Message::data(void) { return zmq_msg_data(&m_message); }

const void *Message::data(void) const
{
	return zmq_msg_data(const_cast<zmq_msg_t *>(&m_message));
}

std::size_t Message::size(void) const
{
	return zmq_msg_size(const_cast<zmq_msg_t *>(&m_message));
}

zmq_msg_t *Message::message(void) { return &m_message; }

const int Socket::RCVTIMEO;

Socket::Socket(void) : m_socket(nullptr) {}

Socket::~Socket(void) { shutdown(); }

bool Socket::create(void *_context, int _type)
{
	if ((m_socket = zmq_socket(_context, _type)) == nullptr) {
		logr::error << "Failed to create socket" << zendl;
		return false;
	}

	int zero = 0;
	int rcvtimeo = RCVTIMEO;

	return setsockopt(ZMQ_LINGER, &zero, sizeof(zero)) &&
	       setsockopt(ZMQ_RCVTIMEO, &rcvtimeo, sizeof(rcvtimeo));
}

void Socket::shutdown(void)
{
	if (m_socket && zmq_close(m_socket) < 0) {
		logr::warning << "Failed to close socket" << zendl;
	}

	m_socket = nullptr;
}

bool Socket::setsockopt(int _name, const void *_value, std::size_t _len)
{
	if (zmq_setsockopt(m_socket, _name, _value, _len) < 0) {
		logr::error << "Failed to configure socket" << zendl;
		return false;
	}

	return true;
}

bool Socket::bind(const std::string &_endpoint)
{
	if (zmq_bind(m_socket, _endpoint.c_str()) < 0) {
		logr::error << "Failed to bind socket: " << _endpoint << zendl;
		return false;
	}

	return true;
}

bool Socket::connect(const std::string &_endpoint)
{
	if (zmq_connect(m_socket, _endpoint.c_str()) < 0) {
		logr::error << "Failed to connect socket: " << _endpoint
			    << zendl;
		return false;
	}

	return true;
}

bool Socket::send(Message &_message)
{
	if (zmq_msg_send(_message.message(), m_socket, 0) < 0) {
		if (!running()) {
			return false;
		}

		logr::error << "Failed to send message" << zendl;
		return false;
	}

	return true;
}

bool Socket::recv(Message &_message)
{
	while (true) {
		if (zmq_msg_recv(_message.message(), m_socket, 0) < 0) {
			int error = zmq_errno();

			if (!running() || error == ETERM) {
				return false;
			}

			if (error == EAGAIN) {
				continue;
			}

			logr::error << "Failed to receive message" << zendl;
			return false;
		}

		return true;
	}
}

bool Socket::running(void) const { return m_socket; }
}

API::API(const Config &_config, const Statistics &_statistics)
	: m_config(_config), m_statistics(_statistics), m_context(nullptr)
{
	GOOGLE_PROTOBUF_VERIFY_VERSION;

	if ((m_context = zmq::zmq_ctx_new()) == nullptr) {
		logr::error << "Failed to create context" << zmq::zendl;
		throw std::runtime_error(zmq::zmq_strerror(zmq::zmq_errno()));
	}
}

API::~API(void)
{
	google::protobuf::ShutdownProtobufLibrary();
	shutdown();

	if (zmq::zmq_ctx_term(m_context) < 0) {
		logr::warning << "Failed to terminate context" << zmq::zendl;
	}
}

bool API::start(void)
{
	if (!m_control.create(m_context, ZMQ_REP)) {
		shutdown();
		return false;
	}

	if (!m_downlink.create(m_context, ZMQ_PUSH)) {
		shutdown();
		return false;
	}

	if (!m_uplink.create(m_context, ZMQ_SUB)) {
		shutdown();
		return false;
	}

#ifdef DEPRECATED_FEATURES
	if (!m_lorad.create(m_context, ZMQ_REQ)) {
		shutdown();
		return false;
	}

	int one = 1;

	if (!m_lorad.setsockopt(ZMQ_REQ_RELAXED, &one, sizeof(one))) {
		shutdown();
		return false;
	}
#endif

	for (const auto &control : m_config.apiControl) {
		if (!m_control.bind(control)) {
			shutdown();
			return false;
		}
	}

	for (const auto &downlink : m_config.apiDownlink) {
		if (!m_downlink.connect(downlink)) {
			shutdown();
			return false;
		}
	}

	for (const auto &uplink : m_config.apiUplink) {
		if (!m_uplink.connect(uplink)) {
			shutdown();
			return false;
		}
	}

#ifdef DEPRECATED_FEATURES
	if (!m_lorad.connect(m_config.gwmpLoradControl)) {
		shutdown();
		return false;
	}
#endif

	for (const auto &filter : m_config.apiFilters) {
		if (!subscribe(filter)) {
			shutdown();
			return false;
		}
	}

	// Be sure to always subscribe to the events.
	if (!subscribe("event")) {
		shutdown();
		return false;
	}

	if (!add_workers(std::bind(&API::rpc, this))) {
		shutdown();
		return false;
	}

	return true;
}

void API::shutdown(void)
{
	Service::shutdown();

#ifdef DEPRECATED_FEATURES
	m_lorad.shutdown();
#endif
	m_uplink.shutdown();
	m_downlink.shutdown();
	m_control.shutdown();

#if ZMQ_VERSION >= ZMQ_MAKE_VERSION(4, 2, 3)
	std::this_thread::sleep_for(
		std::chrono::milliseconds(zmq::Socket::RCVTIMEO));
#endif
}

bool API::running(void) const
{
	return Service::running() && m_context && m_control.running() &&
	       m_downlink.running() && m_uplink.running()
#ifdef DEPRECATED_FEATURES
	       && m_lorad.running()
#endif
		;
}

// The configuration filters for the uplink channel can handle raw binary
// buffers (e.g "\xdeadbeef") or keywords.
//
// Keywords are case-insensitive and are one of these:
// - "lora"
// - "gfsk" or "fsk"
// - "event"
//
// Keywords have to be identified and then translated into raw binary
// corresponding to the protocol buffer messages.
//
// This namespace contains all the helpers needed to handle this point.
namespace {
std::string to_keyword(const std::string &_filter)
{
	std::string filter;

	std::transform(_filter.begin(),
		       _filter.end(),
		       std::back_inserter(filter),
		       ::tolower);

	if (filter == "fsk") {
		return "gfsk";
	}

	return filter;
}

std::string binaryLora(void)
{
	proto::uplink::Container proto;
	proto.mutable_lora()->set_message_id(0);
	return proto.SerializeAsString().substr(0, 1);
}

std::string binaryGfsk(void)
{
	proto::uplink::Container proto;
	proto.mutable_gfsk()->set_message_id(0);
	return proto.SerializeAsString().substr(0, 1);
}

std::string binaryEvent(void)
{
	proto::uplink::Container proto;
	proto.mutable_event()->mutable_system()->set_event(
		static_cast<proto::event::System::Event>(0));
	return proto.SerializeAsString().substr(0, 1);
}

std::string to_binary(const std::string &_filter)
{
	std::string filter = to_keyword(_filter);

	if (filter == "lora") {
		return binaryLora();

	} else if (filter == "gfsk") {
		return binaryGfsk();

	} else if (filter == "event") {
		return binaryEvent();
	}

	return _filter;
}
}

bool API::subscribe(const std::string &_filter)
{
	std::string filter = to_binary(_filter);

	return subscribe(filter.data(), filter.size());
}

bool API::subscribe(const char *_filter, std::size_t _size)
{
	return m_uplink.setsockopt(ZMQ_SUBSCRIBE, _filter, _size);
}

// This namespace contains all the "message to protobuf" helpers.
namespace {
template <typename MessageType, typename ProtoType>
ProtoType to_proto(const MessageType &_message, ProtoType)
{
	return static_cast<ProtoType>(_message);
}

#define TO_PROTO(_message, _proto, _name)                                      \
	(_proto)->set_##_name(to_proto((_message)->_name, (_proto)->_name()))

struct TimingVisitor
{
	proto::downlink::Timing *m_proto;

	TimingVisitor(proto::downlink::Timing *_proto) : m_proto(_proto) {}

	bool operator()(std::uint32_t _counter)
	{
		m_proto->set_counter(_counter);

		return true;
	}

	bool operator()(const message::time::AccurateContinuousTime &_time)
	{
		auto *proto = m_proto->mutable_time();

		TO_PROTO(&_time, proto, format);
		TO_PROTO(&_time, proto, time);
		TO_PROTO(&_time, proto, microseconds);

		return true;
	}

	bool operator()(const message::downlink::TimeBoundaries &_boundaries)
	{
		auto *proto = m_proto->mutable_boundaries();

		if (_boundaries.has_from_counter) {
			TO_PROTO(&_boundaries, proto, from_counter);
		}

		if (_boundaries.has_to_counter) {
			TO_PROTO(&_boundaries, proto, to_counter);
		}

		return true;
	}
};

bool to_proto(const message::downlink::Timing &_message,
	      proto::downlink::Timing *_proto)
{
	TO_PROTO(&_message, _proto, transmission_id);
	TO_PROTO(&_message, _proto, priority);
	TO_PROTO(&_message, _proto, pre_delay);
	TO_PROTO(&_message, _proto, post_delay);

	return mpark::visit(TimingVisitor(_proto), _message.timing);
}

bool to_proto(const message::lora::Configuration &_message,
	      proto::lora::Downlink::Configuration *_proto)
{
	TO_PROTO(&_message, _proto, radio_id);
	TO_PROTO(&_message, _proto, frequency);
	TO_PROTO(&_message, _proto, bandwidth);
	TO_PROTO(&_message, _proto, spreading_factor);
	TO_PROTO(&_message, _proto, coding_rate);

	if (_message.has_power) {
		TO_PROTO(&_message, _proto, power);
	}

	for (const auto &message_timing : _message.timings) {
		auto *proto_timing = _proto->add_timings();

		if (!to_proto(message_timing, proto_timing)) {
			return false;
		}
	}

	return true;
}

bool to_proto(const message::lora::Downlink &_message,
	      proto::downlink::Container &_proto)
{
	auto *proto = _proto.mutable_lora();

	TO_PROTO(&_message, proto, message_id);
	TO_PROTO(&_message, proto, payload);
	TO_PROTO(&_message, proto, required_transmissions);
	TO_PROTO(&_message, proto, preamble);
	TO_PROTO(&_message, proto, crc_enabled);
	TO_PROTO(&_message, proto, header_disabled);
	TO_PROTO(&_message, proto, polarity_inverted);

	for (const auto &message_configuration : _message.configurations) {
		auto *proto_configuration = proto->add_configurations();

		if (!to_proto(message_configuration, proto_configuration)) {
			return false;
		}
	}

	return true;
}

bool to_proto(const message::gfsk::Configuration &_message,
	      proto::gfsk::Downlink::Configuration *_proto)
{
	TO_PROTO(&_message, _proto, radio_id);
	TO_PROTO(&_message, _proto, frequency);
	TO_PROTO(&_message, _proto, bitrate);
	TO_PROTO(&_message, _proto, frequency_deviation);

	if (_message.has_power) {
		TO_PROTO(&_message, _proto, power);
	}

	for (const auto &message_timing : _message.timings) {
		auto *proto_timing = _proto->add_timings();

		if (!to_proto(message_timing, proto_timing)) {
			return false;
		}
	}

	return true;
}

bool to_proto(const message::gfsk::Downlink &_message,
	      proto::downlink::Container &_proto)
{
	auto *proto = _proto.mutable_gfsk();

	TO_PROTO(&_message, proto, message_id);
	TO_PROTO(&_message, proto, payload);
	TO_PROTO(&_message, proto, required_transmissions);
	TO_PROTO(&_message, proto, preamble);
	TO_PROTO(&_message, proto, length_fixed);
	TO_PROTO(&_message, proto, crc);
	TO_PROTO(&_message, proto, encoding);

	for (const auto &message_configuration : _message.configurations) {
		auto *proto_configuration = proto->add_configurations();

		if (!to_proto(message_configuration, proto_configuration)) {
			return false;
		}
	}

	return true;
}

#undef TO_PROTO

struct DownlinkVisitor
{
	proto::downlink::Container &m_proto;

	DownlinkVisitor(proto::downlink::Container &_proto) : m_proto(_proto) {}

	bool operator()(...)
	{
		logr::error << "Unsupported message" << logr::endl;
		return false;
	}

	template <typename T,
		  typename = utils::enable_if_t<
			  utils::is_one_of<T,
					   message::lora::Downlink,
					   message::gfsk::Downlink>::value>>
	bool operator()(const T &_message)
	{
		return to_proto(_message, m_proto);
	}
};

bool to_proto(const API::downlink_type &_message,
	      proto::downlink::Container &_proto)
{
	return mpark::visit(DownlinkVisitor(_proto), _message.messages);
}
}

bool API::sendDownlink(const downlink_type &_message)
{
	proto::downlink::Container proto;
	std::string serialized;

	if (!to_proto(_message, proto)) {
		return false;
	}

	if (!proto.SerializeToString(&serialized)) {
		logr::error << "Failed to serialize message" << logr::endl;
		return false;
	}

	logr::debug << logr::dump(serialized, "api.send") << logr::endl;

	return send(m_downlink, serialized);
}

// This namespace contains all the "protobuf to message" helpers.
namespace {
template <typename ProtoType, typename MessageType>
MessageType to_message(const ProtoType &_proto, MessageType)
{
	return static_cast<MessageType>(_proto);
}

#define TO_MESSAGE(_proto, _message, _name)                                    \
	(_message)->_name = to_message((_proto)->_name(), (_message)->_name)

#define TO_OPTIONAL_MESSAGE(_proto, _message, _name)                           \
	if ((_proto)->has_##_name()) {                                         \
		(_message)->has_##_name = true;                                \
		TO_MESSAGE(_proto, _message, _name);                           \
	}

#define OPTIONAL_BLOCK(_proto, _message, _name, _block)                        \
	if ((_proto)->has_##_name()) {                                         \
		(_message)->has_##_name = true;                                \
		auto &m##_name = (_message)->_name;                            \
		auto &p##_name = (_proto)->_name();                            \
		_block;                                                        \
	}

bool to_message(const proto::lora::Uplink &_proto, API::uplink_type &_message)
{
	auto now = std::chrono::duration_cast<message::recv_time_type>(
		message::clock_type::now().time_since_epoch());

	message::lora::Uplink message{};

	TO_MESSAGE(&_proto, &message, message_id);

	OPTIONAL_BLOCK(&_proto, &message, data, {
		mdata.recv_time = now;

		TO_MESSAGE(&pdata, &mdata, payload);
		TO_OPTIONAL_MESSAGE(&pdata, &mdata, frequency);
		TO_MESSAGE(&pdata, &mdata, crc);
		TO_MESSAGE(&pdata, &mdata, bandwidth);
		TO_MESSAGE(&pdata, &mdata, spreading_factor);
		TO_MESSAGE(&pdata, &mdata, coding_rate);
	});

	for (auto i = 0; i < _proto.radios_size(); i++) {
		message::lora::Radio mradio{};
		auto &pradio = _proto.radios(i);

		TO_MESSAGE(&pradio, &mradio, radio_id);

		OPTIONAL_BLOCK(&pradio, &mradio, demodulator, {
			mdemodulator.recv_time = now;

			TO_MESSAGE(&pdemodulator, &mdemodulator, counter);
			TO_OPTIONAL_MESSAGE(&pdemodulator, &mdemodulator, rssi);

			OPTIONAL_BLOCK(&pdemodulator, &mdemodulator, time, {
				TO_MESSAGE(&ptime, &mtime, format);
				TO_MESSAGE(&ptime, &mtime, time);
				TO_OPTIONAL_MESSAGE(
					&ptime, &mtime, microseconds);
			});

			OPTIONAL_BLOCK(&pdemodulator, &mdemodulator, snr, {
				TO_MESSAGE(&psnr, &msnr, snr);
				TO_OPTIONAL_MESSAGE(&psnr, &msnr, max);
				TO_OPTIONAL_MESSAGE(&psnr, &msnr, min);
			});
		});

		OPTIONAL_BLOCK(&pradio, &mradio, localization, {
			mlocalization.recv_time = now;

			if (plocalization.has_encrypted()) {
				mlocalization.has_nanoseconds = true;
				mlocalization.nanoseconds =
					plocalization.encrypted();

			} else if (plocalization.has_decrypted()) {
				mlocalization.has_nanoseconds = true;
				mlocalization.nanoseconds =
					plocalization.decrypted();

			} else {
				mlocalization.has_nanoseconds = false;
			}

			TO_OPTIONAL_MESSAGE(
				&plocalization, &mlocalization, rssi);
			TO_OPTIONAL_MESSAGE(&plocalization,
					    &mlocalization,
					    standard_deviation);
			TO_OPTIONAL_MESSAGE(&plocalization,
					    &mlocalization,
					    frequency_offset);
			TO_OPTIONAL_MESSAGE(
				&plocalization, &mlocalization, status);
			TO_OPTIONAL_MESSAGE(
				&plocalization, &mlocalization, version);
			TO_OPTIONAL_MESSAGE(&plocalization,
					    &mlocalization,
					    delta_alternate);
		});

		message.radios.push_back(mradio);
	}

	_message.messages = message;

	return true;
}

bool to_message(const proto::gfsk::Uplink &_proto, API::uplink_type &_message)
{
	auto now = std::chrono::duration_cast<message::recv_time_type>(
		message::clock_type::now().time_since_epoch());

	message::gfsk::Uplink message{};

	TO_MESSAGE(&_proto, &message, message_id);

	OPTIONAL_BLOCK(&_proto, &message, data, {
		mdata.recv_time = now;

		TO_MESSAGE(&pdata, &mdata, payload);
		TO_OPTIONAL_MESSAGE(&pdata, &mdata, frequency);
		TO_MESSAGE(&pdata, &mdata, crc);
		TO_MESSAGE(&pdata, &mdata, bandwidth);
		TO_MESSAGE(&pdata, &mdata, bitrate);
	});

	for (auto i = 0; i < _proto.radios_size(); i++) {
		message::gfsk::Radio mradio{};
		auto &pradio = _proto.radios(i);

		TO_MESSAGE(&pradio, &mradio, radio_id);

		OPTIONAL_BLOCK(&pradio, &mradio, demodulator, {
			mdemodulator.recv_time = now;

			TO_MESSAGE(&pdemodulator, &mdemodulator, counter);
			TO_OPTIONAL_MESSAGE(&pdemodulator, &mdemodulator, rssi);

			OPTIONAL_BLOCK(&pdemodulator, &mdemodulator, time, {
				TO_MESSAGE(&ptime, &mtime, format);
				TO_MESSAGE(&ptime, &mtime, time);
				TO_OPTIONAL_MESSAGE(
					&ptime, &mtime, microseconds);
			});
		});

		message.radios.push_back(mradio);
	}

	_message.messages = message;

	return true;
}

bool to_message(const proto::event::System &_proto,
		message::event::Uplink &_message)
{
	message::event::System message{};

	TO_MESSAGE(&_proto, &message, event);

	_message.events = message;

	return true;
}

bool to_message(const proto::event::Transmission &_proto,
		message::event::Uplink &_message)
{
	message::event::Transmission message{};

	TO_MESSAGE(&_proto, &message, message_id);
	TO_MESSAGE(&_proto, &message, transmission_id);
	TO_MESSAGE(&_proto, &message, status);
	TO_OPTIONAL_MESSAGE(&_proto, &message, radio_id);
	TO_MESSAGE(&_proto, &message, pre_delay);
	TO_MESSAGE(&_proto, &message, post_delay);
	TO_OPTIONAL_MESSAGE(&_proto, &message, power);

	OPTIONAL_BLOCK(&_proto, &message, time, {
		TO_MESSAGE(&ptime, &mtime, format);
		TO_MESSAGE(&ptime, &mtime, time);
		TO_OPTIONAL_MESSAGE(&ptime, &mtime, microseconds);
	});

	_message.events = message;

	return true;
}

bool to_message(const proto::event::Uplink &_proto, API::uplink_type &_message)
{
	message::event::Uplink message{};

	if (_proto.has_system()) {
		if (!to_message(_proto.system(), message)) {
			return false;
		}

	} else if (_proto.has_transmission()) {
		if (!to_message(_proto.transmission(), message)) {
			return false;
		}

	} else {
		logr::error << "Unknown event message" << logr::endl;
		return false;
	}

	_message.messages = message;

	return true;
}

#undef OPTIONAL_BLOCK
#undef TO_MESSAGE
#undef TO_OPTIONAL_MESSAGE
}

bool API::recvUplink(uplink_type &_message)
{
	proto::uplink::Container proto;
	std::string serialized;

	while (running()) {
		if (!recv(m_uplink, serialized)) {
			return false;
		}

		logr::debug << logr::dump(serialized, "api.recv") << logr::endl;

		if (!proto.ParseFromString(serialized)) {
			logr::error << "Failed to de-serialize message"
				    << logr::endl;
			return false;
		}

		switch (proto.messages_case()) {
		case proto::uplink::Container::kLora:
			return to_message(proto.lora(), _message);

		case proto::uplink::Container::kGfsk:
			return to_message(proto.gfsk(), _message);

		case proto::uplink::Container::kEvent:
			return to_message(proto.event(), _message);

		case proto::uplink::Container::kScan:
			logr::debug << "Dropping scan message" << logr::endl;
			continue;

		default:
			logr::error << "Unknown uplink message" << logr::endl;
			return false;
		}
	}

	return false;
}

bool API::send(zmq::Socket &_socket, const std::string &_message)
{
	zmq::Message message;

	if (!message.assign(_message.data(), _message.size())) {
		return false;
	}

	return _socket.send(message);
}

bool API::recv(zmq::Socket &_socket, std::string &_message)
{
	zmq::Message message;

	if (!_socket.recv(message)) {
		return false;
	}

	_message = std::string(
		static_cast<const std::string::value_type *>(message.data()),
		message.size());

	return true;
}

namespace {
proto::control::Reply rpc_set_verbosity(const proto::control::Request &_request)
{
	proto::verbosity::Request protobuf{};
	proto::control::Reply reply{};

	if (!protobuf.ParseFromString(_request.protobuf())) {
		reply.set_status(proto::control::Reply::BAD_REQUEST);

	} else {
		logr::setLevel(static_cast<logr::level>(protobuf.level()));
		reply.set_status(proto::control::Reply::OK);
	}

	return reply;
}

proto::control::Reply rpc_get_version(void)
{
	proto::version::Reply protobuf{};
	proto::control::Reply reply{};
	std::string serialized;

	protobuf.set_version("lorafwd " VERSION);

	if (!protobuf.SerializeToString(&serialized)) {
		reply.set_status(proto::control::Reply::FAILED);
	} else {
		reply.set_protobuf(serialized);
		reply.set_status(proto::control::Reply::OK);
	}

	return reply;
}

#define add_statistic_uint(reply, stat, name)                                  \
	{                                                                      \
		auto *statistic = (reply)->add_statistics();                   \
		statistic->set_name(#name);                                    \
		statistic->set_unsigned_int((stat)->name);                     \
	}

proto::control::Reply rpc_get_statistics(const Statistics &_stat)
{
	proto::statistic::Reply protobuf{};
	proto::control::Reply reply{};
	std::string serialized;

	add_statistic_uint(&protobuf, &_stat, uplink_forwarded);
	add_statistic_uint(&protobuf, &_stat, uplink_sent);
	add_statistic_uint(&protobuf, &_stat, uplink_acknowledged);
	add_statistic_uint(&protobuf, &_stat, heartbeat_sent);
	add_statistic_uint(&protobuf, &_stat, heartbeat_acknowledged);
	add_statistic_uint(&protobuf, &_stat, downlink_received);
	add_statistic_uint(&protobuf, &_stat, downlink_acknowledged);

	auto *time = protobuf.mutable_time();

	time->set_format(proto::time::Time::UNIX_TIME);
	time->set_time(std::time(nullptr));

	if (!protobuf.SerializeToString(&serialized)) {
		reply.set_status(proto::control::Reply::FAILED);

	} else {
		reply.set_protobuf(serialized);
		reply.set_status(proto::control::Reply::OK);
	}

	return reply;
}

#undef add_statistic_uint
}

bool API::rpc(void)
{
	proto::control::Request request{};
	proto::control::Reply reply{};

	std::string serialized;

	if (!recv(m_control, serialized)) {
		if (!running()) {
			return false;
		}

		logr::error << "Failed to receive control message"
			    << logr::endl;
		return true;
	}

	logr::debug << logr::dump(serialized, "api.req") << logr::endl;

	if (!request.ParseFromString(serialized)) {
		logr::error << "Failed to de-serialize message" << logr::endl;
		return true;
	}

	if (request.service() == "set_verbosity") {
		reply = rpc_set_verbosity(request);

	} else if (request.service() == "get_version") {
		reply = rpc_get_version();

	} else if (request.service() == "get_statistics") {
		reply = rpc_get_statistics(m_statistics);

	} else {
		reply.set_status(proto::control::Reply::NOT_IMPLEMENTED);
	}

	if (!reply.SerializeToString(&serialized)) {
		logr::error << "Failed to serialize message" << logr::endl;
		return true;
	}

	logr::debug << logr::dump(serialized, "api.rep") << logr::endl;

	if (!send(m_control, serialized)) {
		if (!running()) {
			return false;
		}

		logr::error << "Failed to send control message" << logr::endl;
		return true;
	}

	return true;
}

#ifdef DEPRECATED_FEATURES
#define add_statistic_int(stat, proto, namestr)                                \
	{                                                                      \
		if (#namestr == proto.name()) {                                \
			stat.namestr = proto.signed_int();                     \
			stat.has_##namestr = true;                             \
		}                                                              \
	}

#define add_statistic_uint(stat, proto, namestr)                               \
	{                                                                      \
		if (#namestr == proto.name()) {                                \
			stat.namestr = proto.unsigned_int();                   \
			stat.has_##namestr = true;                             \
		}                                                              \
	}

#define add_statistic_float(stat, proto, namestr)                              \
	{                                                                      \
		if (#namestr == proto.name()) {                                \
			stat.namestr = proto.float_();                         \
			stat.has_##namestr = true;                             \
		}                                                              \
	}

#define add_statistic_str(stat, proto, namestr)                                \
	{                                                                      \
		if (#namestr == proto.name()) {                                \
			stat.namestr = proto.string();                         \
			stat.has_##namestr = true;                             \
		}                                                              \
	}

bool API::requestStatistics(LoradStatistics &_statistics)
{
	proto::statistic::Reply protobuf{};
	proto::control::Request request{};
	proto::control::Reply reply{};
	std::string serialized;

	request.set_service("get_statistics");

	if (!request.SerializeToString(&serialized)) {
		logr::error << "Failed to serialize message" << logr::endl;
		return false;
	}

	if (!send(m_lorad, serialized)) {
		if (running()) {
			logr::error << "Failed to send lorad message"
				    << logr::endl;
		}

		return false;
	}

	if (!recv(m_lorad, serialized)) {
		if (running()) {
			logr::error << "Failed to receive lorad message"
				    << logr::endl;
		}

		return false;
	}

	if (!reply.ParseFromString(serialized)) {
		logr::error << "Failed to de-serialize message" << logr::endl;
		return false;
	}

	if (reply.status() != proto::control::Reply::OK) {
		logr::error << "Invalid reply status" << logr::endl;
		return false;
	}

	if (!reply.has_protobuf() ||
	    !protobuf.ParseFromString(reply.protobuf())) {
		logr::error << "Failed to de-serialize message" << logr::endl;
		return false;
	}

	for (auto i = 0; i < protobuf.statistics_size(); i++) {
		auto &stat = protobuf.statistics(i);

		add_statistic_uint(_statistics, stat, uplink_received);
		add_statistic_uint(_statistics, stat, uplink_invalid);
		add_statistic_uint(_statistics, stat, uplink_valid);
		add_statistic_uint(_statistics, stat, uplink_bytes_received);
		add_statistic_uint(_statistics, stat, downlink_requested);
		add_statistic_uint(_statistics, stat, downlink_scheduled);
		add_statistic_uint(_statistics, stat, downlink_transmitted);
		add_statistic_uint(
			_statistics, stat, downlink_bytes_transmitted);
		add_statistic_float(_statistics, stat, gps_latitude);
		add_statistic_float(_statistics, stat, gps_longitude);
		add_statistic_int(_statistics, stat, gps_altitude);
		add_statistic_uint(_statistics, stat, gps_lost_pulse);
		add_statistic_float(_statistics, stat, radio_temperature);
		add_statistic_uint(_statistics, stat, version_fpga);
		add_statistic_uint(_statistics, stat, version_dsp);
		add_statistic_str(_statistics, stat, version_hal);
	}

	return true;
}

#undef add_statistic_uint
#endif
