// Copyright (c) 2018 Kerlink SA. All rights reserved.

#include "database.h"

#include <iterator>
#include <ostream>
#include <sstream>
#include <stdexcept>
#include <string>

#include "third-party/variant.h"

#include "utils/logger.h"
#include "utils/type_traits.h"

#ifndef VARSTATEDIR
#define VARSTATEDIR "/var/lib/lorafwd"
#endif

namespace {
struct DBendl
{
	sqlite3pp::database &db;
};

DBendl dbendl(sqlite3pp::database &_db) { return {_db}; }

std::ostream &operator<<(std::ostream &_os, const DBendl &_db)
{
	return (_os << ": " << _db.db.error_msg() << std::endl);
}
}

const int Database::SQLite::MODULATION_LORA;
const int Database::SQLite::MODULATION_GFSK;

Database::SQLite::SQLite(const std::string &_filename) : m_filename(_filename)
{
}

bool Database::SQLite::open(void)
{
	int rc;

	rc = m_db.connect(m_filename.c_str(),
			  SQLITE_OPEN_READWRITE | SQLITE_OPEN_CREATE);

	if (rc != SQLITE_OK) {
		logr::error << "Failed to connect database" << dbendl(m_db);
		return false;
	}

	if (m_db.enable_foreign_keys() != SQLITE_OK) {
		logr::error << "Failed to enable foreign keys" << dbendl(m_db);
		return false;
	}

	constexpr int BUSY_TIMEOUT = 200;

	if (m_db.set_busy_timeout(BUSY_TIMEOUT) != SQLITE_OK) {
		logr::error << "Failed to set busy timeout" << dbendl(m_db);
		return false;
	}

	rc = m_db.execute("CREATE TABLE IF NOT EXISTS versions ("
			  " database INTEGER NOT NULL,"
			  " binary TEXT NOT NULL UNIQUE)");

	if (rc != SQLITE_OK) {
		logr::error << "Failed to create table: versions"
			    << dbendl(m_db);
		return false;
	}

	try {
		std::size_t size = 0;
		std::string version;
		int dbversion;

		sqlite3pp::query query(m_db,
				       "SELECT database, binary"
				       " FROM versions"
				       " ORDER BY rowid DESC LIMIT 1");

		for (auto it = query.begin(); it != query.end(); ++it, ++size) {
			dbversion = (*it).get<int>(0);
			version = (*it).get<std::string>(1);

			logr::info << "Database version: " << dbversion << " ("
				   << version << ")" << logr::endl;
		}

		if (size == 0) {
			if (!create()) {
				return false;
			}

		} else {
			if (dbversion > DBVERSION) {
				logr::error
					<< "Database version not supported: "
					<< dbversion << " (" << version << ")"
					<< logr::endl;
				return false;
			}

			if (dbversion < DBVERSION && !update(dbversion)) {
				return false;
			}
		}

		sqlite3pp::command command(m_db,
					   "INSERT INTO versions("
					   " database, binary)"
					   " VALUES (?, ?)");

		command.bind(1, DBVERSION);
		command.bind(2, VERSION, sqlite3pp::copy);

		rc = command.execute();

		if (rc != SQLITE_OK && rc != SQLITE_CONSTRAINT) {
			logr::error << "Failed to insert database version"
				    << dbendl(m_db);
			return false;
		}

		rc = m_db.execute("UPDATE datas SET recv_time = 0");

		if (rc != SQLITE_OK) {
			logr::error << "Failed to reset recv_time: datas"
				    << dbendl(m_db);
			return false;
		}

		rc = m_db.execute("UPDATE demodulators SET recv_time = 0");

		if (rc != SQLITE_OK) {
			logr::error << "Failed to reset recv_time: demodulators"
				    << dbendl(m_db);
			return false;
		}

		rc = m_db.execute("UPDATE localizations SET recv_time = 0");

		if (rc != SQLITE_OK) {
			logr::error
				<< "Failed to reset recv_time: localizations"
				<< dbendl(m_db);
			return false;
		}

		return true;

	} catch (const std::exception &e) {
		logr::error << "Failed to check version: " << e.what()
			    << logr::endl;
		return false;
	}
}

bool Database::SQLite::create(void)
{
	int rc;

	rc = m_db.execute("CREATE TABLE IF NOT EXISTS messages ("
			  " id INTEGER PRIMARY KEY AUTOINCREMENT,"
			  " message_id INTEGER NOT NULL UNIQUE)");

	if (rc != SQLITE_OK) {
		logr::error << "Failed to create table: messages"
			    << dbendl(m_db);
		return false;
	}

	rc = m_db.execute("CREATE TABLE IF NOT EXISTS datas ("
			  " message INTEGER NOT NULL UNIQUE,"
			  " recv_time INTEGER NOT NULL DEFAULT 0,"
			  " modulation INTEGER NOT NULL,"
			  " payload BLOB NOT NULL,"
			  " frequency INTEGER,"
			  " crc INTEGER,"
			  " bandwidth INTEGER,"
			  " spreading_factor INTEGER,"
			  " coding_rate INTEGER,"
			  " bitrate INTEGER,"
			  " FOREIGN KEY(message)"
			  "  REFERENCES messages(id)"
			  "  ON DELETE CASCADE)");

	if (rc != SQLITE_OK) {
		logr::error << "Failed to create table: datas" << dbendl(m_db);
		return false;
	}

	rc = m_db.execute("CREATE TABLE IF NOT EXISTS radios ("
			  " id INTEGER PRIMARY KEY AUTOINCREMENT,"
			  " message INTEGER NOT NULL,"
			  " radio_id INTEGER NOT NULL,"
			  " UNIQUE (message, radio_id) ON CONFLICT ABORT,"
			  " FOREIGN KEY(message)"
			  "  REFERENCES messages(id)"
			  "  ON DELETE CASCADE)");

	if (rc != SQLITE_OK) {
		logr::error << "Failed to create table: radios" << dbendl(m_db);
		return false;
	}

	rc = m_db.execute("CREATE TABLE IF NOT EXISTS demodulators ("
			  " radio INTEGER NOT NULL UNIQUE,"
			  " recv_time INTEGER NOT NULL DEFAULT 0,"
			  " counter INTEGER NOT NULL,"
			  " time_format INTEGER,"
			  " time INTEGER,"
			  " microseconds INTEGER,"
			  " rssi REAL,"
			  " snr REAL,"
			  " snr_max REAL,"
			  " snr_min REAL,"
			  " FOREIGN KEY(radio)"
			  "  REFERENCES radios(id)"
			  "  ON DELETE CASCADE)");

	if (rc != SQLITE_OK) {
		logr::error << "Failed to create table: demodulators"
			    << dbendl(m_db);
		return false;
	}

	rc = m_db.execute("CREATE TABLE IF NOT EXISTS localizations ("
			  " radio INTEGER NOT NULL UNIQUE,"
			  " recv_time INTEGER NOT NULL DEFAULT 0,"
			  " nanoseconds_encrypted BLOB,"
			  " nanoseconds_decrypted INTEGER,"
			  " rssi REAL,"
			  " standard_deviation INTEGER,"
			  " frequency_offset INTEGER,"
			  " status INTEGER,"
			  " version INTEGER,"
			  " delta_alternate INTEGER,"
			  " FOREIGN KEY(radio)"
			  "  REFERENCES radios(id)"
			  "  ON DELETE CASCADE)");

	if (rc != SQLITE_OK) {
		logr::error << "Failed to create table: localizations"
			    << dbendl(m_db);
		return false;
	}

	return true;
}

bool Database::SQLite::update(int from)
{
	int rc;

	logr::notice << "Updating database version from " << from << " to "
		     << DBVERSION << logr::endl;

	switch (from) {
	case 1:
		rc = m_db.execute("ALTER TABLE datas"
				  " ADD COLUMN"
				  "  recv_time INTEGER NOT NULL DEFAULT 0");

		if (rc != SQLITE_OK) {
			logr::error << "Failed to update table: datas"
				    << dbendl(m_db);
			return false;
		}

		rc = m_db.execute("ALTER TABLE demodulators"
				  " ADD COLUMN"
				  "  recv_time INTEGER NOT NULL DEFAULT 0");

		if (rc != SQLITE_OK) {
			logr::error << "Failed to update table: demodulators"
				    << dbendl(m_db);
			return false;
		}

		rc = m_db.execute("ALTER TABLE localizations"
				  " ADD COLUMN"
				  "  recv_time INTEGER NOT NULL DEFAULT 0");

		if (rc != SQLITE_OK) {
			logr::error << "Failed to update table: localizations"
				    << dbendl(m_db);
			return false;
		}

	/* fallthrough */

	default:
		return true;
	}
}

namespace {
template <typename T>
typename utils::enable_if_t<
	utils::disjunction<utils::is_int<T>,
			   utils::is_one_of<typename std::remove_cv<T>::type,
					    message::gfsk::Bandwidth,
					    message::lora::Bandwidth,
					    message::lora::CodingRate,
					    message::lora::SpreadingFactor,
					    message::time::TimeFormat,
					    message::uplink::CRCStatus>>::value>
sqlbind(sqlite3pp::statement &_stmt, int _index, T _value)
{
	_stmt.bind(_index, static_cast<int>(_value));
}

void sqlbind(sqlite3pp::statement &_stmt,
	     int _index,
	     message::recv_time_type _value)
{
	_stmt.bind(_index, static_cast<long long int>(_value.count()));
}

template <typename T>
typename utils::enable_if_t<utils::is_float<T>::value>
sqlbind(sqlite3pp::statement &_stmt, int _index, T _value)
{
	_stmt.bind(_index, static_cast<double>(_value));
}

void sqlbind(sqlite3pp::statement &_stmt, int _index, const std::string &_value)
{
	_stmt.bind(_index, _value.data(), _value.size(), sqlite3pp::nocopy);
}

bool is_null(const sqlite3pp::query::rows &_row, int _index)
{
	return _row.column_type(_index) == SQLITE_NULL;
}

template <typename T>
typename utils::enable_if_t<
	utils::disjunction<utils::is_int<T>,
			   utils::is_one_of<typename std::remove_cv<T>::type,
					    message::gfsk::Bandwidth,
					    message::lora::Bandwidth,
					    message::lora::CodingRate,
					    message::lora::SpreadingFactor,
					    message::time::TimeFormat,
					    message::uplink::CRCStatus>>::value,
	T>
to_message(const sqlite3pp::query::rows &_row, int _index, T)
{
	return static_cast<T>(_row.get<int>(_index));
}

message::recv_time_type to_message(const sqlite3pp::query::rows &_row,
				   int _index,
				   message::recv_time_type)
{
	return message::recv_time_type(_row.get<long long int>(_index));
}

template <typename T>
typename utils::enable_if_t<utils::is_float<T>::value, T>
to_message(const sqlite3pp::query::rows &_row, int _index, T)
{
	return static_cast<T>(_row.get<double>(_index));
}

std::string
to_message(const sqlite3pp::query::rows &_row, int _index, std::string)
{
	return std::string(
		static_cast<const char *>(_row.get<const void *>(_index)),
		_row.column_bytes(_index));
}

#define TO_MESSAGE(_row, _index, _parent, _name)                               \
	_parent._name = to_message(_row, _index, _parent._name)

#define TO_OPTIONAL_MESSAGE(_row, _index, _parent, _name)                      \
	if (!is_null(_row, _index)) {                                          \
		_parent.has_##_name = true;                                    \
		TO_MESSAGE(_row, _index, _parent, _name);                      \
	}

#define OPTIONAL_BLOCK(_row, _index, _parent, _name, _block)                   \
	if (!is_null(_row, _index)) {                                          \
		_parent.has_##_name = true;                                    \
		auto &_name = _parent._name;                                   \
		_block;                                                        \
	}

bool to_message(sqlite3pp::database &_db,
		const sqlite3pp::query::rows &_row,
		message::lora::Uplink &_message)
{
	TO_MESSAGE(_row, 1, _message, message_id);

	OPTIONAL_BLOCK(_row, 3, _message, data, {
		TO_MESSAGE(_row, 3, data, payload);
		TO_OPTIONAL_MESSAGE(_row, 4, data, frequency);
		TO_MESSAGE(_row, 5, data, crc);
		TO_MESSAGE(_row, 6, data, bandwidth);
		TO_MESSAGE(_row, 7, data, spreading_factor);
		TO_MESSAGE(_row, 8, data, coding_rate);
		TO_MESSAGE(_row, 10, data, recv_time);
	});

	sqlite3pp::query query(
		_db,
		"SELECT"
		" radio_id,"
		" demodulators.recv_time AS demodulator_recv_time,"
		" counter, time_format, time, microseconds,"
		" demodulators.rssi AS demodulator_rssi,"
		" snr, snr_max, snr_min,"
		" localizations.recv_time AS localization_recv_time,"
		" nanoseconds_encrypted, nanoseconds_decrypted,"
		" localizations.rssi AS localization_rssi,"
		" standard_deviation, frequency_offset, status,"
		" version, delta_alternate"
		" FROM radios"
		" LEFT JOIN demodulators"
		"  ON demodulators.radio = radios.id"
		" LEFT JOIN localizations"
		"  ON localizations.radio = radios.id"
		" WHERE radios.message = ?");

	sqlbind(query, 1, _row.get<int>(0));

	for (auto row : query) {
		message::lora::Radio radio{};

		TO_MESSAGE(row, 0, radio, radio_id);

		OPTIONAL_BLOCK(row, 1, radio, demodulator, {
			TO_MESSAGE(row, 1, demodulator, recv_time);
			TO_MESSAGE(row, 2, demodulator, counter);
			TO_OPTIONAL_MESSAGE(row, 6, demodulator, rssi);

			OPTIONAL_BLOCK(row, 3, demodulator, time, {
				TO_MESSAGE(row, 3, time, format);
				TO_MESSAGE(row, 4, time, time);
				TO_OPTIONAL_MESSAGE(row, 5, time, microseconds);
			});

			OPTIONAL_BLOCK(row, 7, demodulator, snr, {
				TO_MESSAGE(row, 7, snr, snr);
				TO_OPTIONAL_MESSAGE(row, 8, snr, max);
				TO_OPTIONAL_MESSAGE(row, 9, snr, min);
			});
		});

		OPTIONAL_BLOCK(row, 10, radio, localization, {
			TO_MESSAGE(row, 10, localization, recv_time);
			if (!is_null(row, 11)) {
				std::string encrypted;
				encrypted = to_message(row, 11, encrypted);
				localization.has_nanoseconds = true;
				localization.nanoseconds = encrypted;

			} else if (!is_null(row, 12)) {
				message::uint_type decrypted;
				decrypted = to_message(row, 12, decrypted);
				localization.has_nanoseconds = true;
				localization.nanoseconds = decrypted;

			} else {
				localization.has_nanoseconds = false;
			}

			TO_OPTIONAL_MESSAGE(row, 13, localization, rssi);
			TO_OPTIONAL_MESSAGE(
				row, 14, localization, standard_deviation);
			TO_OPTIONAL_MESSAGE(
				row, 15, localization, frequency_offset);
			TO_OPTIONAL_MESSAGE(row, 16, localization, status);
			TO_OPTIONAL_MESSAGE(row, 17, localization, version);
			TO_OPTIONAL_MESSAGE(
				row, 18, localization, delta_alternate);
		});

		_message.radios.push_back(radio);
	}

	return true;
}

bool to_message(sqlite3pp::database &_db,
		const sqlite3pp::query::rows &_row,
		message::gfsk::Uplink &_message)
{
	TO_MESSAGE(_row, 1, _message, message_id);

	OPTIONAL_BLOCK(_row, 3, _message, data, {
		TO_MESSAGE(_row, 3, data, payload);
		TO_OPTIONAL_MESSAGE(_row, 4, data, frequency);
		TO_MESSAGE(_row, 5, data, crc);
		TO_MESSAGE(_row, 6, data, bandwidth);
		TO_MESSAGE(_row, 9, data, bitrate);
		TO_MESSAGE(_row, 10, data, recv_time);
	});

	sqlite3pp::query query(_db,
			       "SELECT"
			       " radio_id, recv_time,"
			       " counter, time_format, time, microseconds,"
			       " rssi"
			       " FROM radios"
			       " LEFT JOIN demodulators"
			       "  ON demodulators.radio = radios.id"
			       " WHERE radios.message = ?");

	sqlbind(query, 1, _row.get<int>(0));

	for (auto row : query) {
		message::gfsk::Radio radio{};

		TO_MESSAGE(row, 0, radio, radio_id);

		OPTIONAL_BLOCK(row, 1, radio, demodulator, {
			TO_MESSAGE(row, 1, demodulator, recv_time);
			TO_MESSAGE(row, 2, demodulator, counter);
			TO_OPTIONAL_MESSAGE(row, 6, demodulator, rssi);

			OPTIONAL_BLOCK(row, 3, demodulator, time, {
				TO_MESSAGE(row, 3, time, format);
				TO_MESSAGE(row, 4, time, time);
				TO_OPTIONAL_MESSAGE(row, 5, time, microseconds);
			});
		});

		_message.radios.push_back(radio);
	}

	return true;
}

#undef OPTIONAL_BLOCK
#undef TO_MESSAGE
#undef TO_OPTIONAL_MESSAGE
}

bool Database::SQLite::fetch(uplinks_type &_messages, std::size_t _size)
{
	_messages.clear();

	try {
		sqlite3pp::query query(m_db,
				       "SELECT"
				       " id, message_id,"
				       " modulation, payload,"
				       " frequency, crc, bandwidth,"
				       " spreading_factor, coding_rate,"
				       " bitrate, recv_time"
				       " FROM messages"
				       " LEFT JOIN datas"
				       "  ON datas.message = messages.id"
				       " ORDER BY id ASC"
				       " LIMIT ?");

		sqlbind(query, 1, _size);

		for (auto row : query) {
			if (row.get<int>(2) == MODULATION_LORA) {
				message::lora::Uplink message{};

				if (!to_message(m_db, row, message)) {
					return false;
				}

				uplink_type container = message;
				_messages.push_back(container);

			} else if (row.get<int>(2) == MODULATION_GFSK) {
				message::gfsk::Uplink message{};

				if (!to_message(m_db, row, message)) {
					return false;
				}

				uplink_type container = message;
				_messages.push_back(container);

			} else {
				logr::error << "Invalid modulation"
					    << logr::endl;
				erase(row.get<int>(1));
			}
		}

		return true;

	} catch (const std::exception &e) {
		logr::error << "Failed to fetch messages: " << e.what()
			    << logr::endl;
		return 0;
	}
}

std::size_t Database::SQLite::size(void)
{
	try {
		int size = 0;

		sqlite3pp::query query(m_db, "SELECT count(*) FROM messages");

		for (auto q : query) {
			q.getter() >> size;
		}

		return size;

	} catch (const std::exception &e) {
		logr::error << "Failed to get message count: " << e.what()
			    << logr::endl;
		return 0;
	}
}

bool Database::SQLite::clear(void)
{
	int rc;

	rc = m_db.execute("DELETE FROM messages");

	if (rc != SQLITE_OK) {
		logr::error << "Failed to delete all messages" << dbendl(m_db);
		return false;
	}

	return true;
}

bool Database::SQLite::erase(id_type _id)
{
	ids_type ids;
	ids.push_back(_id);
	return erase(ids);
}

bool Database::SQLite::erase(const ids_type &_ids)
{
	try {
		std::ostringstream oss;

		oss << "DELETE FROM messages WHERE message_id IN (";

		for (auto id = _ids.begin(); id != _ids.end(); ++id) {
			oss << *id;

			if (std::next(id) != _ids.end()) {
				oss << ", ";
			}
		}

		oss << ")";

		sqlite3pp::command command(m_db, oss.str().c_str());

		if (command.execute() != SQLITE_OK) {
			logr::error << "Failed to erase messages"
				    << dbendl(m_db);
			return false;
		}

		return true;

	} catch (const std::exception &e) {
		logr::error << "Failed to erase messages: " << e.what()
			    << logr::endl;
		return false;
	}
}

bool Database::SQLite::pop_back(void)
{
	int rc;

	rc = m_db.execute("DELETE FROM messages WHERE rowid IN ("
			  " SELECT rowid FROM messages"
			  " ORDER BY id ASC LIMIT 1)");

	if (rc != SQLITE_OK) {
		logr::error << "Failed to delete the last message"
			    << dbendl(m_db);
		return false;
	}

	return true;
}

namespace {
struct LocalizationVisitor
{
	sqlite3pp::command &m_command;

	LocalizationVisitor(sqlite3pp::command &_command) : m_command(_command)
	{
	}

	void operator()(const std::string &_encrypted)
	{
		sqlbind(m_command, 2, _encrypted);
	}

	void operator()(message::uint_type _decrypted)
	{
		sqlbind(m_command, 3, _decrypted);
	}
};

bool to_db(sqlite3pp::database &_db, const message::lora::Uplink &_message)
{
	int rc;

	sqlite3pp::transaction transaction(_db);

	sqlite3pp::command command(_db,
				   "INSERT INTO messages ("
				   " message_id)"
				   " VALUES (?)");

	sqlbind(command, 1, _message.message_id);

	rc = command.execute();

	if (rc != SQLITE_OK && rc != SQLITE_CONSTRAINT) {
		logr::error << "Failed to insert message" << dbendl(_db);
		return false;
	}

	std::size_t size = 0;
	int message_uid;

	sqlite3pp::query query(_db,
			       "SELECT id FROM messages"
			       " WHERE message_id = ?");

	sqlbind(query, 1, _message.message_id);

	for (auto it = query.begin(); it != query.end(); ++it, ++size) {
		(*it).getter() >> message_uid;
	}

	if (size == 0) {
		logr::error << "Failed to query: message uid" << std::endl;
		return false;
	}

	if (_message.has_data) {
		auto &data = _message.data;

		sqlite3pp::command command(
			_db,
			"INSERT INTO datas ("
			" message, modulation, payload,"
			" frequency, crc, bandwidth,"
			" spreading_factor, coding_rate,"
			" recv_time)"
			" VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?)");

		sqlbind(command, 1, message_uid);
		sqlbind(command, 2, Database::SQLite::MODULATION_LORA);
		sqlbind(command, 3, data.payload);

		if (data.has_frequency) {
			sqlbind(command, 4, data.frequency);
		}

		sqlbind(command, 5, data.crc);
		sqlbind(command, 6, data.bandwidth);
		sqlbind(command, 7, data.spreading_factor);
		sqlbind(command, 8, data.coding_rate);
		sqlbind(command, 9, data.recv_time);

		rc = command.execute();

		if (rc != SQLITE_OK && rc != SQLITE_CONSTRAINT) {
			logr::error << "Failed to insert data" << dbendl(_db);
			return false;
		}
	}

	for (const auto &radio : _message.radios) {
		sqlite3pp::command command(_db,
					   "INSERT INTO radios ("
					   " message, radio_id)"
					   " VALUES (?, ?)");

		sqlbind(command, 1, message_uid);
		sqlbind(command, 2, radio.radio_id);

		rc = command.execute();

		if (rc != SQLITE_OK && rc != SQLITE_CONSTRAINT) {
			logr::error << "Failed to insert radio" << dbendl(_db);
			return false;
		}

		std::size_t size = 0;
		int radio_uid;

		sqlite3pp::query query(_db,
				       "SELECT id FROM radios"
				       " WHERE message = ? AND radio_id = ?");

		sqlbind(query, 1, message_uid);
		sqlbind(query, 2, radio.radio_id);

		for (auto it = query.begin(); it != query.end(); ++it, ++size) {
			(*it).getter() >> radio_uid;
		}

		if (size == 0) {
			logr::error << "Failed to query: radio uid"
				    << std::endl;
			return false;
		}

		if (radio.has_demodulator) {
			auto &demodulator = radio.demodulator;

			sqlite3pp::command command(
				_db,
				"INSERT INTO demodulators ("
				" radio, counter, time_format,"
				" time, microseconds, rssi,"
				" snr, snr_max, snr_min,"
				" recv_time)"
				" VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?)");

			sqlbind(command, 1, radio_uid);
			sqlbind(command, 2, demodulator.counter);
			sqlbind(command, 10, demodulator.recv_time);

			if (demodulator.has_time) {
				auto &time = demodulator.time;

				sqlbind(command, 3, time.format);
				sqlbind(command, 4, time.time);

				if (time.has_microseconds) {
					sqlbind(command, 5, time.microseconds);
				}
			}

			if (demodulator.has_rssi) {
				sqlbind(command, 6, demodulator.rssi);
			}

			if (demodulator.has_snr) {
				auto &snr = demodulator.snr;

				sqlbind(command, 7, snr.snr);

				if (snr.has_max) {
					sqlbind(command, 8, snr.max);
				}

				if (snr.has_min) {
					sqlbind(command, 9, snr.min);
				}
			}

			rc = command.execute();

			if (rc != SQLITE_OK && rc != SQLITE_CONSTRAINT) {
				logr::error << "Failed to insert "
					       "demodulator"
					    << dbendl(_db);
				return false;
			}
		}

		if (radio.has_localization) {
			auto &localization = radio.localization;

			sqlite3pp::command command(
				_db,
				"INSERT INTO localizations ("
				" radio, nanoseconds_encrypted,"
				" nanoseconds_decrypted, rssi,"
				" standard_deviation, frequency_offset,"
				" status, version, delta_alternate,"
				" recv_time)"
				" VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?)");

			sqlbind(command, 1, radio_uid);
			sqlbind(command, 10, localization.recv_time);

			if (localization.has_nanoseconds) {
				mpark::visit(LocalizationVisitor(command),
					     localization.nanoseconds);
			}

			if (localization.has_rssi) {
				sqlbind(command, 4, localization.rssi);
			}

			if (localization.has_standard_deviation) {
				sqlbind(command,
					5,
					localization.standard_deviation);
			}

			if (localization.has_frequency_offset) {
				sqlbind(command,
					6,
					localization.frequency_offset);
			}

			if (localization.has_status) {
				sqlbind(command, 7, localization.status);
			}

			if (localization.has_version) {
				sqlbind(command, 8, localization.version);
			}

			if (localization.has_delta_alternate) {
				sqlbind(command,
					9,
					localization.delta_alternate);
			}

			rc = command.execute();

			if (rc != SQLITE_OK && rc != SQLITE_CONSTRAINT) {
				logr::error << "Failed to insert "
					       "localization"
					    << dbendl(_db);
				return false;
			}
		}
	}

	if (transaction.commit() != SQLITE_OK) {
		logr::error << "Failed to commit transaction" << dbendl(_db);
		return false;
	}

	return true;
}

bool to_db(sqlite3pp::database &_db, const message::gfsk::Uplink &_message)
{
	int rc;

	sqlite3pp::transaction transaction(_db);

	sqlite3pp::command command(_db,
				   "INSERT INTO messages ("
				   " message_id)"
				   " VALUES (?)");

	sqlbind(command, 1, _message.message_id);

	rc = command.execute();

	if (rc != SQLITE_OK && rc != SQLITE_CONSTRAINT) {
		logr::error << "Failed to insert message" << dbendl(_db);
		return false;
	}

	std::size_t size = 0;
	int message_uid;

	sqlite3pp::query query(_db,
			       "SELECT id FROM messages"
			       " WHERE message_id = ?");

	sqlbind(query, 1, _message.message_id);

	for (auto it = query.begin(); it != query.end(); ++it, ++size) {
		(*it).getter() >> message_uid;
	}

	if (size == 0) {
		logr::error << "Failed to query: message uid" << std::endl;
		return false;
	}

	if (_message.has_data) {
		auto &data = _message.data;

		sqlite3pp::command command(_db,
					   "INSERT INTO datas ("
					   " message, modulation, payload,"
					   " frequency, crc, bandwidth,"
					   " bitrate, recv_time)"
					   " VALUES (?, ?, ?, ?, ?, ?, ?, ?)");

		sqlbind(command, 1, message_uid);
		sqlbind(command, 2, Database::SQLite::MODULATION_GFSK);
		sqlbind(command, 3, data.payload);

		if (data.has_frequency) {
			sqlbind(command, 4, data.frequency);
		}

		sqlbind(command, 5, data.crc);
		sqlbind(command, 6, data.bandwidth);
		sqlbind(command, 7, data.bitrate);
		sqlbind(command, 8, data.recv_time);

		rc = command.execute();

		if (rc != SQLITE_OK && rc != SQLITE_CONSTRAINT) {
			logr::error << "Failed to insert data" << dbendl(_db);
			return false;
		}
	}

	for (const auto &radio : _message.radios) {
		sqlite3pp::command command(_db,
					   "INSERT INTO radios ("
					   " message, radio_id)"
					   " VALUES (?, ?)");

		sqlbind(command, 1, message_uid);
		sqlbind(command, 2, radio.radio_id);

		rc = command.execute();

		if (rc != SQLITE_OK && rc != SQLITE_CONSTRAINT) {
			logr::error << "Failed to insert radio" << dbendl(_db);
			return false;
		}

		std::size_t size = 0;
		int radio_uid;

		sqlite3pp::query query(_db,
				       "SELECT id FROM radios"
				       " WHERE message = ? AND radio_id = ?");

		sqlbind(query, 1, message_uid);
		sqlbind(query, 2, radio.radio_id);

		for (auto it = query.begin(); it != query.end(); ++it, ++size) {
			(*it).getter() >> radio_uid;
		}

		if (size == 0) {
			logr::error << "Failed to query: radio uid"
				    << std::endl;
			return false;
		}

		if (radio.has_demodulator) {
			auto &demodulator = radio.demodulator;

			sqlite3pp::command command(
				_db,
				"INSERT INTO demodulators ("
				" radio, counter, time_format,"
				" time, microseconds, rssi,"
				" recv_time)"
				" VALUES (?, ?, ?, ?, ?, ?, ?)");

			sqlbind(command, 1, radio_uid);
			sqlbind(command, 2, demodulator.counter);
			sqlbind(command, 7, demodulator.recv_time);

			if (demodulator.has_time) {
				auto &time = demodulator.time;

				sqlbind(command, 3, time.format);
				sqlbind(command, 4, time.time);

				if (time.has_microseconds) {
					sqlbind(command, 5, time.microseconds);
				}
			}

			if (demodulator.has_rssi) {
				sqlbind(command, 6, demodulator.rssi);
			}

			rc = command.execute();

			if (rc != SQLITE_OK && rc != SQLITE_CONSTRAINT) {
				logr::error << "Failed to insert "
					       "demodulator"
					    << dbendl(_db);
				return false;
			}
		}
	}

	if (transaction.commit() != SQLITE_OK) {
		logr::error << "Failed to commit transaction" << dbendl(_db);
		return false;
	}

	return true;
}

struct UplinkVisitor
{
	sqlite3pp::database &m_db;

	UplinkVisitor(sqlite3pp::database &_db) : m_db(_db) {}

	bool operator()(...)
	{
		logr::error << "Unsupported message" << logr::endl;
		return false;
	}

	template <typename T,
		  typename = utils::enable_if_t<
			  utils::is_one_of<T,
					   message::lora::Uplink,
					   message::gfsk::Uplink>::value>>
	bool operator()(const T &_message)
	{
		try {
			return to_db(m_db, _message);

		} catch (const std::exception &e) {
			logr::error << "Failed to push message: " << e.what()
				    << logr::endl;
			return false;
		}
	}
};
}

bool Database::SQLite::push_front(const uplink_type &_message)
{
	return mpark::visit(UplinkVisitor(m_db), _message.messages);
}

Database::Database(const Config &_config)
	: m_config(_config), m_sqlite(VARSTATEDIR "/database.sqlite"), m_size(0)
{
}

bool Database::start(void)
{
	{
		std::lock_guard<std::mutex> lock(m_mutex);

		if (!m_sqlite.open()) {
			return false;
		}

		if (!m_config.dbEnable) {
			return m_sqlite.clear();
		}

		m_size = m_sqlite.size();

		m_running = true;
	}

	if (m_size) {
		m_not_empty.notify_one();
	}

	return true;
}

void Database::shutdown(void)
{
	Service::shutdown();
	m_not_empty.notify_all();
}

bool Database::insert(const uplink_type &_message)
{
	if (!m_config.dbEnable) {
		return false;
	}

	{
		std::lock_guard<std::mutex> lock(m_mutex);

		if (!m_sqlite.push_front(_message)) {
			return false;
		}

		m_size = m_sqlite.size();

		if (m_size > m_config.dbLimitMessages) {
			logr::warning << "Database message limit reached"
				      << logr::endl;

			while (m_size > m_config.dbLimitMessages) {
				logr::info << "Dropping the last message"
					   << logr::endl;

				if (!m_sqlite.pop_back()) {
					return false;
				}

				m_size--;
			}
		}
	}

	m_not_empty.notify_one();

	return true;
}

bool Database::fetch(uplinks_type &_messages, std::size_t _size)
{
	if (!m_config.dbEnable) {
		return false;
	}

	std::unique_lock<std::mutex> lock(m_mutex);

	m_not_empty.wait(lock, [&] { return m_size || !running(); });

	if (!running()) {
		return false;
	}

	return m_sqlite.fetch(_messages, _size);
}

bool Database::erase(const ids_type &_ids)
{
	if (!m_config.dbEnable) {
		return false;
	}

	std::lock_guard<std::mutex> lock(m_mutex);

	if (!m_sqlite.erase(_ids)) {
		return false;
	}

	m_size = m_sqlite.size();

	return true;
}
