// Copyright (c) 2018 Kerlink SA. All rights reserved.

#include "forwarder.h"

#include "third-party/variant.h"

#include "utils/logger.h"
#include "utils/type_traits.h"

#include "message/container.h"

Forwarder::Forwarder(const Config &_config,
		     API &_api,
		     Filter &_filter,
		     Database &_database,
		     GWMP &_gwmp)
	: m_config(_config), m_api(_api), m_filter(_filter),
	  m_database(_database), m_gwmp(_gwmp)
{
	m_dbLimitMessages = m_config.dbLimitMessages;
	m_dbCooldown = m_config.dbDelayCooldown;

	m_gwmp.registerHandler(std::bind(&Forwarder::acknowledge, this),
			       std::bind(&Forwarder::timeout, this));
}

bool Forwarder::start(void)
{
	return add_workers(std::bind(&Forwarder::database, this),
			   std::bind(&Forwarder::downlink, this),
			   std::bind(&Forwarder::uplink, this));
}

bool Forwarder::database(void)
{
	std::vector<message::uplink::Container> messages;

	if (!m_config.dbEnable) {
		return false;
	}

	if (!m_database.fetch(messages, m_dbLimitMessages)) {
		if (!m_database.running()) {
			return false;
		}

		logr::error << "Failed to fetch uplink messages" << logr::endl;
		return true;
	}

	if (!m_gwmp.sendUplink(messages)) {
		if (!m_gwmp.running()) {
			return false;
		}

		logr::error << "Failed to send uplink messages" << logr::endl;
		return true;
	}

	if (!sleep_for(m_dbCooldown)) {
		return false;
	}

	return true;
}

bool Forwarder::downlink(void)
{
	message::downlink::Container message{};

	if (!m_gwmp.recvDownlink(message)) {
		if (!m_gwmp.running()) {
			return false;
		}

		logr::error << "Failed to receive downlink message"
			    << logr::endl;
		return true;
	}

	logr::info << "Received downlink message: " << message << logr::endl;

	if (!m_api.sendDownlink(message)) {
		if (!m_api.running()) {
			return false;
		}

		logr::error << "Failed to send downlink message" << logr::endl;
		return true;
	}

	return true;
}

namespace {
struct EventVisitor
{
	const Config &m_config;
	Database &m_database;
	GWMP &m_gwmp;

	EventVisitor(const Config &_config, Database &_database, GWMP &_gwmp)
		: m_config(_config), m_database(_database), m_gwmp(_gwmp)
	{
	}

	bool operator()(...) { return true; }

	bool operator()(const message::event::Transmission &_message)
	{
		if (!m_gwmp.sendEvent(_message)) {
			if (!m_gwmp.running()) {
				return false;
			}

			logr::error << "Failed to send uplink event"
				    << logr::endl;
			return true;
		}

		return true;
	}
};

struct UplinkVisitor
{
	const Config &m_config;
	Database &m_database;
	GWMP &m_gwmp;

	UplinkVisitor(const Config &_config, Database &_database, GWMP &_gwmp)
		: m_config(_config), m_database(_database), m_gwmp(_gwmp)
	{
	}

	bool operator()(...)
	{
		logr::warning << "Unsupported message" << logr::endl;
		return true;
	}

	template <typename T,
		  typename = utils::enable_if_t<
			  utils::is_one_of<T,
					   message::lora::Uplink,
					   message::gfsk::Uplink>::value>>
	bool operator()(const T &_message)
	{
		if (m_config.dbEnable) {
			if (!m_database.insert(_message)) {
				if (!m_database.running()) {
					return false;
				}

				logr::error << "Failed to insert uplink message"
					    << logr::endl;
				return true;
			}

		} else {
			if (!m_gwmp.sendUplink(_message)) {
				if (!m_gwmp.running()) {
					return false;
				}

				logr::error << "Failed to send uplink messages"
					    << logr::endl;
				return true;
			}
		}

		return true;
	}

	bool operator()(const message::event::Uplink &_message)
	{
		return mpark::visit(EventVisitor(m_config, m_database, m_gwmp),
				    _message.events);
	}
};
}

bool Forwarder::uplink(void)
{
	message::uplink::Container message{};

	if (!m_api.recvUplink(message)) {
		if (!m_api.running()) {
			return false;
		}

		logr::error << "Failed to receive uplink message" << logr::endl;
		return true;
	}

	logr::info << "Received uplink message: " << message << logr::endl;

	if (!m_filter.input(message)) {
		return true;
	}

	return mpark::visit(UplinkVisitor(m_config, m_database, m_gwmp),
			    message.messages);
}

void Forwarder::acknowledge(void)
{
	{
		std::size_t dbLimitMessages = m_dbLimitMessages;

		dbLimitMessages *= 2;

		if (dbLimitMessages > m_config.dbLimitMessages) {
			dbLimitMessages = m_config.dbLimitMessages;
		}

		m_dbLimitMessages = dbLimitMessages;
	}

	{
		std::chrono::milliseconds dbCooldown = m_dbCooldown;

		dbCooldown /= 2;

		if (dbCooldown < m_config.dbDelayCooldown) {
			dbCooldown = m_config.dbDelayCooldown;
		}

		m_dbCooldown = dbCooldown;
	}
}

void Forwarder::timeout(void)
{
	{
		std::size_t dbLimitMessages = m_dbLimitMessages;

		dbLimitMessages /= 2;

		if (dbLimitMessages < 1) {
			dbLimitMessages = 1;
		}

		m_dbLimitMessages = dbLimitMessages;
	}

	{
		std::chrono::milliseconds dbCooldown = m_dbCooldown;

		dbCooldown *= 2;

		if (dbCooldown > GWMP::TIMEOUT) {
			dbCooldown = GWMP::TIMEOUT;
		}

		m_dbCooldown = dbCooldown;
	}
}
