// Copyright (c) 2018 Kerlink SA. All rights reserved.

#include "gwmp.h"

#include <chrono>
#include <cmath>
#include <cstdlib>
#include <cstring>
#include <ctime>
#include <functional>
#include <iomanip>
#include <sstream>
#include <stdexcept>

#include "third-party/json.h"
#include "third-party/variant.h"
using json = nlohmann::json;

#include "utils/logger.h"
#include "utils/stream.h"
#include "utils/type_traits.h"

namespace {
extern "C" {
#include <arpa/inet.h>
#include <netdb.h>
#include <sys/ioctl.h>
#include <sys/select.h>
#include <sys/types.h>
#include <unistd.h>

uint64_t bswap_64(uint64_t _bin)
{
	return ((uint64_t)htonl(_bin & 0xFFFFFFFF) << 32) | htonl(_bin >> 32);
}

uint64_t htonll(uint64_t _host)
{
	return (1 == htonl(1)) ? _host : bswap_64(_host);
}

int wait_for_data(int _sockfd)
{
	struct timeval time;
	fd_set set;

	time.tv_sec = 0;
	time.tv_usec = 200000;

	FD_ZERO(&set);
	FD_SET(_sockfd, &set);

	if (select(_sockfd + 1, &set, nullptr, nullptr, &time) < 0) {
		return -1;
	}

	if (FD_ISSET(_sockfd, &set)) {
		return 1;
	}

	return 0;
}

int get_available_data_size(int _sockfd)
{
	int size;

	if (ioctl(_sockfd, FIONREAD, &size) < 0) {
		return -1;
	}

	return size;
}
}
}

namespace udp {
Socket::Socket(void) : m_sockfd(-1), m_resetting(false) {}

Socket::~Socket(void) { close(); }

bool Socket::create(const char *_node, const char *_service)
{
	addrinfo *infos, *info, hints{};
	int error;

	hints.ai_family = AF_UNSPEC;
	hints.ai_socktype = SOCK_DGRAM;
	hints.ai_protocol = IPPROTO_UDP;
	hints.ai_flags = AI_ADDRCONFIG;

	if ((error = getaddrinfo(_node, _service, &hints, &infos)) != 0) {
		logr::error
			<< "Failed to get address info: " << gai_strerror(error)
			<< logr::endl;
		return false;
	}

	for (info = infos; info != nullptr; info = info->ai_next) {
		if ((m_sockfd = socket(info->ai_family,
				       info->ai_socktype,
				       info->ai_protocol)) < 0) {
			logr::error << "Failed to create socket" << logr::eendl;
			continue;
		}

		m_node = _node;
		m_service = _service;
		m_sockaddrlen = info->ai_addrlen;
		std::memcpy(&m_sockaddr, info->ai_addr, m_sockaddrlen);

		freeaddrinfo(infos);
		return true;
	}

	freeaddrinfo(infos);
	return false;
}

bool Socket::create(const std::string &_node, const std::string &_service)
{
	return create(_node.c_str(), _service.c_str());
}

void Socket::close(void)
{
	if (m_sockfd >= 0) {
		if (::close(m_sockfd) < 0) {
			logr::warning << "Failed to close socket"
				      << logr::eendl;
		}

		m_sockfd = -1;
	}
}

void Socket::shutdown(void)
{
	wait_for_reset();

	close();
}

bool Socket::reset(void)
{
	bool status;

	{
		std::unique_lock<std::mutex> lock(m_mutex);

		m_resetting = true;

		close();
		status = create(m_node.data(), m_service.data());

		m_resetting = false;
	}

	m_condition.notify_all();

	return status;
}

void Socket::wait_for_reset(void)
{
	if (m_resetting) {
		std::unique_lock<std::mutex> lock(m_mutex);

		m_condition.wait(lock, [&] { return !m_resetting; });
	}
}

bool Socket::running(void) const { return m_resetting || m_sockfd >= 0; }

bool Socket::send(const std::string &_message)
{
	logr::debug << logr::dump(_message, "gwmp.send");

	while (true) {
		wait_for_reset();

		if (sendto(m_sockfd,
			   _message.data(),
			   _message.size(),
			   0,
			   reinterpret_cast<const sockaddr *>(&m_sockaddr),
			   m_sockaddrlen) >= 0) {
			return true;
		}

		if (m_resetting) {
			continue;
		}

		if (!running()) {
			return false;
		}

		logr::error << "Failed to send data" << logr::eendl;
		return false;
	}
}

bool Socket::recv(std::string &_message)
{
	std::vector<char> buffer;
	int size;

	while (true) {
		wait_for_reset();

		for (int has_data = 0; has_data == 0;) {
			has_data = wait_for_data(m_sockfd);

			if (m_resetting) {
				break;
			}

			if (!running()) {
				return false;
			}

			if (has_data < 0) {
				logr::error << "Failed to wait for data"
					    << logr::eendl;
				return false;
			}
		}

		if (m_resetting) {
			continue;
		}

		if ((size = get_available_data_size(m_sockfd)) < 0) {
			logr::error << "Failed to get available data size"
				    << logr::eendl;
			return false;
		}

		if (size == 0) {
			continue;
		}

		try {
			buffer.resize(size);

		} catch (const std::exception &e) {
			logr::error << "Failed to allocate buffer memory: "
				    << e.what() << logr::endl;
			return false;
		}

		if (recvfrom(m_sockfd,
			     buffer.data(),
			     buffer.size(),
			     0,
			     nullptr,
			     nullptr) >= 0) {
			_message = std::string(buffer.cbegin(), buffer.cend());
			logr::debug << logr::dump(_message, "gwmp.recv");
			return true;
		}

		if (m_resetting) {
			continue;
		}

		if (!running()) {
			return false;
		}

		logr::error << "Failed to receive data" << logr::eendl;
		return false;
	}
}
}

constexpr GWMP::timeout_type GWMP::TIMEOUT;

GWMP::Manager::Manager(const std::string &_name)
	: m_name(_name), m_maxPings(3), m_database(nullptr), m_timedOut(0),
	  m_maxTimedOut(0), m_maxTimedOutHandler({})
{
	std::srand(std::time(nullptr));
	m_nextId = std::rand();

	add_ping(duration_type(TIMEOUT).count());
}

void GWMP::Manager::attachDatabase(Database *_database)
{
	m_database = _database;
}

void GWMP::Manager::registerHandler(handler_type _maxTimedOutHandler,
				    std::size_t _maxTimedOut)
{
	m_maxTimedOutHandler = _maxTimedOutHandler;
	m_maxTimedOut = _maxTimedOut;
}

void GWMP::Manager::registerHandler(handler_type _acknowledgedHandler,
				    handler_type _timedOutHandler)
{
	m_acknowledgedHandler = _acknowledgedHandler;
	m_timedOutHandler = _timedOutHandler;
}

GWMP::id_type GWMP::Manager::next_id(void) { return m_nextId++; }

void GWMP::Manager::sent(id_type _id, const db_ids_type &_db_ids)
{
	auto now = clock_type::now();

	{
		std::lock_guard<std::mutex> lock(m_mutex);

		m_history[_id] = {
			.time = now, .db_ids = _db_ids,
		};
	}

	logr::info << m_name << " (" << utils::to_hex(_id) << ") sent"
		   << logr::endl;

	clear();
}

void GWMP::Manager::acknowledged(id_type _id)
{
	auto now = clock_type::now();
	auto timeout = now - TIMEOUT;
	{
		std::lock_guard<std::mutex> lock(m_mutex);

		auto it = m_history.find(_id);

		if (it != m_history.end() && it->second.time >= timeout) {
			duration_type duration = now - it->second.time;
			add_ping(duration.count());

			if (m_database && !it->second.db_ids.empty()) {
				m_database->erase(it->second.db_ids);
			}

			logr::info << m_name << " (" << utils::to_hex(it->first)
				   << ") acknowledged in " << duration.count()
				   << " ms" << logr::endl;

			if (m_maxTimedOut) {
				m_timedOut = 0;
			}

			if (m_acknowledgedHandler) {
				m_acknowledgedHandler();
			}

			m_history.erase(it);
		}
	}

	clear();
}

void GWMP::Manager::clear(void)
{
	auto now = clock_type::now();
	auto timeout = now - TIMEOUT;

	{
		std::lock_guard<std::mutex> lock(m_mutex);

		for (auto it = m_history.cbegin(), next_it = it;
		     it != m_history.cend();
		     it = next_it) {
			++next_it;

			if (it->second.time < timeout) {
				add_ping(TIMEOUT);

				logr::warning << m_name << " ("
					      << utils::to_hex(it->first)
					      << ") timed out" << logr::endl;

				if (m_maxTimedOut) {
					m_timedOut++;
				}

				if (m_timedOutHandler) {
					m_timedOutHandler();
				}

				m_history.erase(it);
			}
		}
	}

	if (m_maxTimedOut && m_timedOut >= m_maxTimedOut) {
		m_timedOut = 0;

		if (m_maxTimedOutHandler) {
			m_maxTimedOutHandler();
		}
	}
}

void GWMP::Manager::add_ping(const ping_type &_ping)
{
	m_pings.push_back(_ping);

	while (m_pings.size() > m_maxPings) {
		m_pings.pop_front();
	}
}

void GWMP::Manager::add_ping(const duration_type &_duration)
{
	add_ping(_duration.count());
}

GWMP::Manager::ping_type GWMP::Manager::ping(void) const
{
	ping_type pings = 0;

	if (m_pings.size() > 0) {
		for (const auto &ping : m_pings) {
			pings += ping;
		}

		pings /= m_pings.size();
	}

	return pings;
}

GWMP::GWMP(const Config &_config,
	   API &_api,
	   Database &_database,
	   Statistics &_statistics)
	: m_config(_config), m_api(_api), m_database(_database),
	  m_statistics(_statistics), m_downlinkManager("Heartbeat"),
	  m_uplinkManager("Uplink message")
{
}

GWMP::~GWMP(void) { shutdown(); }

bool GWMP::start(void)
{
	if (!m_downlinkSocket.create(m_config.gwmpNode,
				     m_config.gwmpServiceDownlink)) {
		logr::error << "Failed to start downlink socket" << logr::endl;
		return false;
	}

	if (!m_uplinkSocket.create(m_config.gwmpNode,
				   m_config.gwmpServiceUplink)) {
		logr::error << "Failed to start uplink socket" << logr::endl;
		return false;
	}

	m_downlinkManager.registerHandler(std::bind(&GWMP::restart, this),
					  m_config.gwmpAutorestart);

	m_uplinkManager.attachDatabase(&m_database);

	return add_workers(std::bind(&GWMP::acknowledger, this),
			   std::bind(&GWMP::heartbeat, this),
			   std::bind(&GWMP::statistics, this),
			   std::bind(&GWMP::timeout, this));
}

void GWMP::restart(void)
{
	logr::info << "Restarting GWMP sockets" << logr::endl;

	if (!m_downlinkSocket.reset()) {
		logr::warning << "Failed to reset downlink socket"
			      << logr::endl;
	}

	if (!m_uplinkSocket.reset()) {
		logr::warning << "Failed to reset uplink socket" << logr::endl;
	}
}

void GWMP::shutdown(void)
{
	Service::shutdown();
	m_downlinkSocket.shutdown();
	m_uplinkSocket.shutdown();
}

bool GWMP::running(void) const
{
	return Service::running() && m_downlinkSocket.running() &&
	       m_uplinkSocket.running();
}

void GWMP::registerHandler(handler_type _acknowledgedHandler,
			   handler_type _timedOutHandler)
{
	m_uplinkManager.registerHandler(_acknowledgedHandler, _timedOutHandler);
}

namespace {
enum class Version : std::uint8_t {
	V1 = 1,
	V2 = 2,
};

enum class MessageType : std::uint8_t {
	PUSH_DATA = 0,
	PUSH_ACK = 1,
	PULL_DATA = 2,
	PULL_RESP = 3,
	PULL_ACK = 4,
	TX_ACK = 5,
};

struct __attribute__((packed)) HeaderSend
{
	Version version;
	GWMP::id_type id;
	MessageType type;
	std::uint64_t gateway_id;
};

struct __attribute__((packed)) HeaderRecv
{
	Version version;
	GWMP::id_type id;
	MessageType type;
};

std::string to_string(const HeaderSend &_header)
{
	HeaderSend header = _header;

	header.id = htons(header.id);
	header.gateway_id = htonll(header.gateway_id);

	return std::string(reinterpret_cast<const char *>(&header),
			   sizeof(header));
}

HeaderRecv to_header(const std::string &_string)
{
	HeaderRecv header;

	std::memcpy(&header, _string.data(), sizeof(header));

	header.id = ntohs(header.id);

	return header;
}
}

namespace {
std::string to_base64(const std::string &_src)
{
	std::size_t i, len = _src.size();
	std::string dst;

	const std::string table = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
				  "abcdefghijklmnopqrstuvwxyz"
				  "0123456789+/";

	for (i = 0; (len > 2) && (i < len - 2); i += 3) {
		dst += table[((_src[i + 0] >> 2) & 0x3F)];
		dst += table[((_src[i + 0] << 4) & 0x30) |
			     ((_src[i + 1] >> 4) & 0x0F)];
		dst += table[((_src[i + 1] << 2) & 0x3C) |
			     ((_src[i + 2] >> 6) & 0x03)];
		dst += table[((_src[i + 2] >> 0) & 0x3F)];
	}

	if (i < len) {
		dst += table[((_src[i + 0] >> 2) & 0x3F)];

		if (i == len - 1) {
			dst += table[((_src[i + 0] << 4) & 0x30)];
			dst += "=";

		} else {
			dst += table[((_src[i + 0] << 4) & 0x30) |
				     ((_src[i + 1] >> 4) & 0x0F)];
			dst += table[((_src[i + 1] << 2) & 0x3C)];
		}

		dst += "=";
	}

	return dst;
}

std::string from_base64(const std::string &_src)
{
	std::size_t i, len;
	std::string dst;

	const char table[UINT8_MAX + 1] = {
		64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
		64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
		64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 62, 64, 64, 64, 63,
		52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 64, 64, 64, 64, 64, 64,
		64, 0,  1,  2,  3,  4,  5,  6,  7,  8,  9,  10, 11, 12, 13, 14,
		15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 64, 64, 64, 64, 64,
		64, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40,
		41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 64, 64, 64, 64, 64,
		64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
		64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
		64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
		64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
		64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
		64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
		64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
		64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64};

	for (len = 0; table[static_cast<std::uint8_t>(_src[len])] < 64; len++)
		;

	for (i = 0; len - i > 4; i += 4) {
		dst += table[static_cast<std::uint8_t>(_src[i])] << 2 |
		       table[static_cast<std::uint8_t>(_src[i + 1])] >> 4;
		dst += table[static_cast<std::uint8_t>(_src[i + 1])] << 4 |
		       table[static_cast<std::uint8_t>(_src[i + 2])] >> 2;
		dst += table[static_cast<std::uint8_t>(_src[i + 2])] << 6 |
		       table[static_cast<std::uint8_t>(_src[i + 3])];
	}

	if (len - i > 1) {
		dst += table[static_cast<std::uint8_t>(_src[i])] << 2 |
		       table[static_cast<std::uint8_t>(_src[i + 1])] >> 4;
	}

	if (len - i > 2) {
		dst += table[static_cast<std::uint8_t>(_src[i + 1])] << 4 |
		       table[static_cast<std::uint8_t>(_src[i + 2])] >> 2;
	}

	if (len - i > 3) {
		dst += table[static_cast<std::uint8_t>(_src[i + 2])] << 6 |
		       table[static_cast<std::uint8_t>(_src[i + 3])];
	}

	return dst;
}
}

namespace {
template <typename T>
T setprecision(T _value, int _precision)
{
	T integral;
	T fractional = std::modf(_value, &integral);

	std::stringstream string;
	T fractional_reduced;

	string << std::setprecision(_precision) << fractional;
	string >> fractional_reduced;

	return integral + fractional_reduced;
}
}

namespace {
using tmms_type = std::uint64_t;

tmms_type to_tmms(const message::time::Time &_time)
{
	tmms_type tmms = _time.time;

	if (_time.format == message::time::TimeFormat::TAI_TIME) {
		tmms -= message::time::GPS_UNIX_EPOCH +
			message::time::TAI_GPS_OFFSET;
	}

	tmms *= 1000;

	if (_time.has_microseconds) {
		tmms += _time.microseconds / 1000;
	}

	return tmms;
}

message::time::AccurateContinuousTime to_time(tmms_type _tmms)
{
	message::time::AccurateContinuousTime time{};

	time.format = message::time::ContinuousTimeFormat::GPS_TIME;
	time.time = _tmms / 1000;
	time.microseconds = (_tmms % 1000) * 1000;

	return time;
}
}

namespace {
bool to_json(const GWMP::event_type &_message,
	     std::string &_string,
	     GWMP::message_id_type &_message_id)
{
	try {
		json root;

		switch (_message.status) {
		case message::event::Status::TRANSMITTED:
			root["txpk_ack"]["error"] = "NONE";
			break;
		case message::event::Status::BAD_POWER:
			root["txpk_ack"]["error"] = "TX_POWER";
			break;
		case message::event::Status::BAD_FREQUENCY:
			root["txpk_ack"]["error"] = "TX_FREQ";
			break;
		case message::event::Status::BAD_TIMING:
			root["txpk_ack"]["error"] = "TOO_LATE";
			break;
		case message::event::Status::GPS_UNLOCKED:
			root["txpk_ack"]["error"] = "GPS_UNLOCKED";
			break;
		case message::event::Status::SCHEDULING_FAILED:
			root["txpk_ack"]["error"] = "COLLISION_PACKET";
			break;
		case message::event::Status::CHANNEL_BUSY:
			root["txpk_ack"]["error"] = "SEND_LBT";
			break;
		default:
			root["txpk_ack"]["error"] = "UNKNOWN";
			break;
		}

		_message_id = _message.message_id;
		_string = root.dump();

		return true;

	} catch (const std::exception &e) {
		logr::error << "Failed to serialize message: " << e.what()
			    << logr::endl;
		return false;
	}
}
}

bool GWMP::sendEvent(const event_type &_message)
{
	std::string message;
	message_id_type message_id;

	if (!to_json(_message, message, message_id)) {
		return false;
	}

	HeaderSend header = {
		.version = Version::V2,
		.id = static_cast<id_type>(message_id),
		.type = MessageType::TX_ACK,
		.gateway_id = m_config.gatewayId,
	};

	message.insert(0, to_string(header));

	if (!m_downlinkSocket.send(message)) {
		return false;
	}

	m_statistics.downlink_acknowledged++;
	logr::info << "Downlink message (" << utils::to_hex(header.id)
		   << ") acknowledged" << logr::endl;

	return true;
}

namespace {
bool to_message(const std::string &_string,
		message::lora::Bandwidth &_bandwidth,
		message::lora::SpreadingFactor &_spreading_factor)
{
	std::istringstream string(_string);
	std::string sf_str, bw_str;
	std::uint32_t sf, bw;

	string >> std::setw(2) >> sf_str >> sf >> std::setw(2) >> bw_str >> bw;

	if (sf_str != "SF" || bw_str != "BW") {
		logr::error << "Invalid data rate" << logr::endl;
		return false;
	}

	switch (bw) {
	case 125:
		_bandwidth = message::lora::Bandwidth::BW_125K;
		break;
	case 250:
		_bandwidth = message::lora::Bandwidth::BW_250K;
		break;
	case 500:
		_bandwidth = message::lora::Bandwidth::BW_500K;
		break;
	default:
		logr::error << "Invalid bandwidth" << logr::endl;
		return false;
	}

	switch (sf) {
	case 7:
		_spreading_factor = message::lora::SpreadingFactor::SF_7;
		return true;
	case 8:
		_spreading_factor = message::lora::SpreadingFactor::SF_8;
		return true;
	case 9:
		_spreading_factor = message::lora::SpreadingFactor::SF_9;
		return true;
	case 10:
		_spreading_factor = message::lora::SpreadingFactor::SF_10;
		return true;
	case 11:
		_spreading_factor = message::lora::SpreadingFactor::SF_11;
		return true;
	case 12:
		_spreading_factor = message::lora::SpreadingFactor::SF_12;
		return true;
	default:
		logr::error << "Invalid spreading factor" << logr::endl;
		return false;
	}
}

bool to_message(const std::string &_string,
		message::lora::CodingRate &_coding_rate)
{
	if (_string == "4/5") {
		_coding_rate = message::lora::CodingRate::CR_1;
		return true;
	} else if (_string == "4/6" || _string == "2/3") {
		_coding_rate = message::lora::CodingRate::CR_2;
		return true;
	} else if (_string == "4/7") {
		_coding_rate = message::lora::CodingRate::CR_3;
		return true;
	} else if (_string == "4/8" || _string == "2/4" || _string == "1/2") {
		_coding_rate = message::lora::CodingRate::CR_4;
		return true;
	} else {
		logr::error << "Invalid coding rate" << logr::endl;
		return false;
	}
}

bool has(const json &_json, const std::string &_key)
{
	return _json.find(_key) != _json.end() && !_json.at(_key).is_null();
}

template <typename T>
T optional(const json &_json, const std::string &_key, const T &_default)
{
	return (has(_json, _key)) ? _json.at(_key).get<T>() : _default;
}

const message::uint_type PREAMBLE_LORA_DEFAULT = 8;
const message::uint_type PREAMBLE_LORA_MIN = 6;

bool to_message(const json &_json,
		GWMP::message_id_type _message_id,
		message::lora::Downlink &_message)
{
	message::lora::Downlink message{};

	message.message_id = _message_id;
	message.payload = from_base64(_json.at("data"));
	message.required_transmissions = 1;
	message.preamble = optional(_json, "prea", PREAMBLE_LORA_DEFAULT);
	message.crc_enabled = !optional(_json, "ncrc", false);
	message.polarity_inverted = optional(_json, "ipol", false);

	if (_json.at("size") != message.payload.size()) {
		logr::error << "Invalid message size" << logr::endl;
		return false;
	}

	if (message.preamble < PREAMBLE_LORA_MIN) {
		message.preamble = PREAMBLE_LORA_MIN;
	}

	message::lora::Configuration configuration{};

	if (has(_json, "brd")) {
		configuration.radio_id = _json["brd"];

	} else if (has(_json, "ant")) {
		configuration.radio_id = message::radio_id::to_radio_id(
			0, 0, 0, 0, _json["ant"], 0);

	} else if (has(_json, "rfch")) {
		configuration.radio_id = message::radio_id::to_radio_id(
			0, 0, 0, 0, _json["rfch"], 0);
	}

	configuration.frequency =
		_json.at("freq").get<json::number_float_t>() * 1000000;

	if (has(_json, "powe")) {
		configuration.has_power = true;
		configuration.power = _json["powe"];
	}

	if (!to_message(_json.at("datr"),
			configuration.bandwidth,
			configuration.spreading_factor)) {
		return false;
	}

	if (!to_message(_json.at("codr"), configuration.coding_rate)) {
		return false;
	}

	message::downlink::Timing timing{};

	timing.priority = 1;

	if (has(_json, "imme") && _json["imme"] == true) {
		timing.timing = message::downlink::TimeBoundaries{};

	} else if (has(_json, "tmst")) {
		timing.timing = static_cast<message::uint_type>(_json["tmst"]);

	} else if (has(_json, "tmms")) {
		timing.timing = to_time(_json["tmms"]);

	} else {
		logr::error << "Invalid timing" << logr::endl;
		return false;
	}

	configuration.timings.push_back(timing);
	message.configurations.push_back(configuration);
	_message = message;

	return true;
}

const message::uint_type PREAMBLE_GFSK_DEFAULT = 5;
const message::uint_type PREAMBLE_GFSK_MIN = 3;

bool to_message(const json &_json,
		GWMP::message_id_type _message_id,
		message::gfsk::Downlink &_message)
{
	message::gfsk::Downlink message{};

	message.message_id = _message_id;
	message.payload = from_base64(_json.at("data"));
	message.required_transmissions = 1;
	message.preamble = optional(_json, "prea", PREAMBLE_GFSK_DEFAULT);
	message.encoding = message::gfsk::EncodingType::ENC_WHITENING;

	if (_json.at("size") != message.payload.size()) {
		logr::error << "Invalid message size" << logr::endl;
		return false;
	}

	if (!optional(_json, "ncrc", false)) {
		message.crc = message::gfsk::CRCType::CRC_CCITT;
	} else {
		message.crc = message::gfsk::CRCType::CRC_DISABLED;
	}

	if (message.preamble < PREAMBLE_GFSK_MIN) {
		message.preamble = PREAMBLE_GFSK_MIN;
	}

	message::gfsk::Configuration configuration{};

	if (has(_json, "brd")) {
		configuration.radio_id = _json["brd"];

	} else if (has(_json, "ant")) {
		configuration.radio_id = message::radio_id::to_radio_id(
			0, 0, 0, 0, _json["ant"], 0);

	} else if (has(_json, "rfch")) {
		configuration.radio_id = message::radio_id::to_radio_id(
			0, 0, 0, 0, _json["rfch"], 0);
	}

	configuration.frequency =
		_json.at("freq").get<json::number_float_t>() * 1000000;
	configuration.bitrate = _json.at("datr");
	configuration.frequency_deviation = _json.at("fdev");

	if (has(_json, "powe")) {
		configuration.has_power = true;
		configuration.power = _json["powe"];
	}

	message::downlink::Timing timing{};

	timing.priority = 1;

	if (has(_json, "imme") && _json["imme"] == true) {
		timing.timing = message::downlink::TimeBoundaries{};

	} else if (has(_json, "tmst")) {
		timing.timing = static_cast<message::uint_type>(_json["tmst"]);

	} else if (has(_json, "tmms")) {
		timing.timing = to_time(_json["tmms"]);

	} else {
		logr::error << "Invalid timing" << logr::endl;
		return false;
	}

	configuration.timings.push_back(timing);
	message.configurations.push_back(configuration);
	_message = message;

	return true;
}

bool to_message(const std::string &_json,
		GWMP::message_id_type _message_id,
		GWMP::downlink_type &_message)
{
	try {
		json root = json::parse(_json);

		if (root.at("txpk").at("modu") == "LORA") {
			message::lora::Downlink lora{};

			if (!to_message(root.at("txpk"), _message_id, lora)) {
				return false;
			}

			_message.messages = lora;
			return true;

		} else if (root.at("txpk").at("modu") == "FSK") {
			message::gfsk::Downlink gfsk{};

			if (!to_message(root.at("txpk"), _message_id, gfsk)) {
				return false;
			}

			_message.messages = gfsk;
			return true;

		} else {
			logr::error << "Invalid modulation" << logr::endl;
			return false;
		}

	} catch (const std::exception &e) {
		logr::error << "Failed to de-serialize message: " << e.what()
			    << logr::endl;
		return false;
	}
}
}

bool GWMP::recvDownlink(downlink_type &_message)
{
	while (m_downlinkSocket.running()) {
		std::string message;

		if (!m_downlinkSocket.recv(message)) {
			return false;
		}

		HeaderRecv header = to_header(message);

		if (header.type == MessageType::PULL_RESP) {
			if (!to_message(message.substr(sizeof(header)),
					header.id,
					_message)) {
				return false;
			}

			m_statistics.downlink_received++;
			logr::info << "Downlink message ("
				   << utils::to_hex(header.id) << ") received"
				   << logr::endl;

			return true;

		} else if (header.type == MessageType::PULL_ACK) {
			m_downlinkManager.acknowledged(header.id);
			m_statistics.heartbeat_acknowledged++;
		}
	}

	return false;
}

namespace {
int to_integer(message::uplink::CRCStatus _crc)
{
	switch (_crc) {
	case message::uplink::CRCStatus::CRC_OK:
		return 1;
	case message::uplink::CRCStatus::CRC_NONE:
		return 0;
	default:
		return -1;
	}
}

std::string to_string(message::lora::Bandwidth _bandwidth,
		      message::lora::SpreadingFactor _spreading_factor)
{
	std::string datarate;

	switch (_spreading_factor) {
	case message::lora::SpreadingFactor::SF_7:
		datarate += "SF7";
		break;
	case message::lora::SpreadingFactor::SF_8:
		datarate += "SF8";
		break;
	case message::lora::SpreadingFactor::SF_9:
		datarate += "SF9";
		break;
	case message::lora::SpreadingFactor::SF_10:
		datarate += "SF10";
		break;
	case message::lora::SpreadingFactor::SF_11:
		datarate += "SF11";
		break;
	case message::lora::SpreadingFactor::SF_12:
		datarate += "SF12";
		break;
	default:
		return "unknown";
	}

	switch (_bandwidth) {
	case message::lora::Bandwidth::BW_125K:
		datarate += "BW125";
		break;
	case message::lora::Bandwidth::BW_250K:
		datarate += "BW250";
		break;
	case message::lora::Bandwidth::BW_500K:
		datarate += "BW500";
		break;
	default:
		return "unknown";
	}

	return datarate;
}

std::string to_string(message::lora::CodingRate _coding_rate)
{
	switch (_coding_rate) {
	case message::lora::CodingRate::CR_4_5:
		return "4/5";
	case message::lora::CodingRate::CR_4_6:
		return "4/6";
	case message::lora::CodingRate::CR_4_7:
		return "4/7";
	case message::lora::CodingRate::CR_4_8:
		return "4/8";
	default:
		return "unknown";
	}
}

struct LocalizationVisitor
{
	json &m_rsig;

	LocalizationVisitor(json &_rsig) : m_rsig(_rsig) {}

	void operator()(const std::string &_encrypted)
	{
		m_rsig["etime"] = to_base64(_encrypted);
	}

	void operator()(message::uint_type _decrypted)
	{
		m_rsig["ftime"] = _decrypted;
	}
};

bool to_json(const message::lora::Uplink &_message,
	     json &_json,
	     GWMP::db_ids_type &_db_ids)
{
	auto now = std::chrono::duration_cast<message::recv_time_type>(
		message::clock_type::now().time_since_epoch());

	message::recv_time_type recv_time(0);
	message::float_type rssi_max = -1000;

	if (!_message.has_data) {
		logr::error << "Invalid message without data" << logr::endl;
		return false;
	}

	auto &data = _message.data;

	_json["jver"] = 2;
	_json["modu"] = "LORA";

	if (data.crc != message::uplink::CRCStatus::CRC_ERROR) {
		_json["data"] = to_base64(data.payload);
		_json["size"] = data.payload.size();

	} else {
		_json["data"] = "";
		_json["size"] = 0;
	}

	_json["stat"] = to_integer(data.crc);
	_json["datr"] = to_string(data.bandwidth, data.spreading_factor);
	_json["codr"] = to_string(data.coding_rate);
	_json["rsig"] = json::array();

	if (!data.has_frequency) {
		logr::error << "Invalid message without frequency"
			    << logr::endl;
		return false;
	}

	_json["freq"] =
		static_cast<json::number_float_t>(data.frequency) / 1000000;

	if (data.recv_time > recv_time) {
		recv_time = data.recv_time;
	}

	for (const auto &radio : _message.radios) {
		if (!radio.has_demodulator) {
			logr::error << "Invalid message without demodulator"
				    << logr::endl;
			return false;
		}

		auto &demodulator = radio.demodulator;
		json rsig;

		// GWMP store radio related data in the global part of the
		// message. In order to be the most accurate, these data will
		// be picked form the radio with the best RSSI.
		if (demodulator.has_rssi && demodulator.rssi > rssi_max) {
			rssi_max = demodulator.rssi;

			_json["tmst"] = demodulator.counter;
			_json["brd"] = radio.radio_id;
			_json["aesk"] =
				message::radio_id::to_board(radio.radio_id);

			if (demodulator.has_time) {
				auto &time = demodulator.time;

				switch (time.format) {
				case message::time::TimeFormat::UNIX_TIME:
					_json["time"] =
						message::time::to_string(time);
					break;

				default:
					_json["tmms"] = to_tmms(time);
					break;
				}
			}
		}

		rsig["ant"] = message::radio_id::to_antenna(radio.radio_id);
		rsig["chan"] = message::radio_id::to_channel(radio.radio_id);

		if (demodulator.has_rssi) {
			rsig["rssic"] = static_cast<int>(demodulator.rssi);
		}

		if (demodulator.has_snr) {
			rsig["lsnr"] =
				setprecision(static_cast<json::number_float_t>(
						     demodulator.snr.snr),
					     1);
		}

		if (demodulator.recv_time > recv_time) {
			recv_time = demodulator.recv_time;
		}

		if (radio.has_localization) {
			auto &localization = radio.localization;

			if (localization.has_nanoseconds) {
				mpark::visit(LocalizationVisitor(rsig),
					     localization.nanoseconds);
			}

			if (localization.has_rssi) {
				rsig["rssis"] =
					static_cast<int>(localization.rssi);
			}

			if (localization.has_standard_deviation) {
				rsig["rssisd"] =
					localization.standard_deviation;
			}

			if (localization.has_frequency_offset) {
				rsig["foff"] = localization.frequency_offset;
			}

			if (localization.has_status) {
				rsig["ftstat"] = localization.status;
			}

			if (localization.has_version) {
				rsig["ftver"] = localization.version;
			}

			if (localization.has_delta_alternate) {
				rsig["ftdelta"] = localization.delta_alternate;
			}

			if (localization.recv_time > recv_time) {
				recv_time = localization.recv_time;
			}
		}

		_json["rsig"] += rsig;
	}

	if (now - message::RX2 > recv_time) {
		_json["delayed"] = true;
	}

	_db_ids.push_back(_message.message_id);

	return true;
}

bool to_json(const message::gfsk::Uplink &_message,
	     json &_json,
	     GWMP::db_ids_type &_db_ids)
{
	auto now = std::chrono::duration_cast<message::recv_time_type>(
		message::clock_type::now().time_since_epoch());

	message::recv_time_type recv_time(0);
	message::float_type rssi_max = -1000;

	if (!_message.has_data) {
		logr::error << "Invalid message without data" << logr::endl;
		return false;
	}

	auto &data = _message.data;

	_json["jver"] = 2;
	_json["modu"] = "FSK";

	if (data.crc != message::uplink::CRCStatus::CRC_ERROR) {
		_json["data"] = to_base64(data.payload);
		_json["size"] = data.payload.size();

	} else {
		_json["data"] = "";
		_json["size"] = 0;
	}

	_json["stat"] = to_integer(data.crc);
	_json["datr"] = data.bitrate;
	_json["rsig"] = json::array();

	if (!data.has_frequency) {
		logr::error << "Invalid message without frequency"
			    << logr::endl;
		return false;
	}

	_json["freq"] =
		static_cast<json::number_float_t>(data.frequency) / 1000000;

	if (data.recv_time > recv_time) {
		recv_time = data.recv_time;
	}

	for (const auto &radio : _message.radios) {
		if (!radio.has_demodulator) {
			logr::error << "Invalid message without demodulator"
				    << logr::endl;
			return false;
		}

		auto &demodulator = radio.demodulator;
		json rsig;

		// GWMP store radio related data in the global part of the
		// message. In order to be the most accurate, these data will
		// be picked form the radio with the best RSSI.
		if (demodulator.has_rssi && demodulator.rssi > rssi_max) {
			rssi_max = demodulator.rssi;

			_json["tmst"] = demodulator.counter;
			_json["brd"] = radio.radio_id;
			_json["aesk"] =
				message::radio_id::to_board(radio.radio_id);

			if (demodulator.has_time) {
				auto &time = demodulator.time;

				switch (time.format) {
				case message::time::TimeFormat::UNIX_TIME:
					_json["time"] =
						message::time::to_string(time);
					break;

				default:
					_json["tmms"] = to_tmms(time);
					break;
				}
			}
		}

		rsig["ant"] = message::radio_id::to_antenna(radio.radio_id);
		rsig["chan"] = message::radio_id::to_channel(radio.radio_id);

		if (demodulator.has_rssi) {
			rsig["rssic"] = static_cast<int>(demodulator.rssi);
		}

		if (demodulator.recv_time > recv_time) {
			recv_time = demodulator.recv_time;
		}

		_json["rsig"] += rsig;
	}

	if (now - message::RX2 > recv_time) {
		_json["delayed"] = true;
	}

	_db_ids.push_back(_message.message_id);

	return true;
}

struct UplinkVisitor
{
	json &m_json;
	GWMP::db_ids_type &m_db_ids;

	UplinkVisitor(json &_json, GWMP::db_ids_type &_db_ids)
		: m_json(_json), m_db_ids(_db_ids)
	{
	}

	bool operator()(...)
	{
		logr::error << "Unsupported message" << logr::endl;
		return false;
	}

	template <typename T,
		  typename = utils::enable_if_t<
			  utils::is_one_of<T,
					   message::lora::Uplink,
					   message::gfsk::Uplink>::value>>
	bool operator()(const T &_message)
	{
		return to_json(_message, m_json, m_db_ids);
	}
};

bool to_json(const GWMP::uplink_type &_message,
	     json &_json,
	     GWMP::db_ids_type &_db_ids)
{
	return mpark::visit(UplinkVisitor(_json, _db_ids), _message.messages);
}

bool to_json(const GWMP::uplinks_type &_messages,
	     std::string &_json,
	     GWMP::db_ids_type &_db_ids)
{
	json root;

	_db_ids.clear();

	root["rxpk"] = json::array();

	for (const auto &message : _messages) {
		json rxpk;

		if (!to_json(message, rxpk, _db_ids)) {
			return false;
		}

		root["rxpk"] += rxpk;
	}

	_json = root.dump();

	return true;
}
}

bool GWMP::sendUplink(const std::string &_message, const db_ids_type &_db_ids)
{
	std::string message = _message;

	HeaderSend header = {
		.version = Version::V2,
		.id = m_uplinkManager.next_id(),
		.type = MessageType::PUSH_DATA,
		.gateway_id = m_config.gatewayId,
	};

	message.insert(0, to_string(header));

	if (!m_uplinkSocket.send(message)) {
		return false;
	}

	m_uplinkManager.sent(header.id, _db_ids);
	m_statistics.uplink_forwarded++;
	m_statistics.uplink_sent++;

	return true;
}

bool GWMP::sendUplink(const uplink_type &_message)
{
	uplinks_type messages;

	messages.push_back(_message);

	return sendUplink(messages);
}

bool GWMP::sendUplink(const uplinks_type &_messages)
{
	try {
		uplinks_type messages = _messages;
		std::string message;
		db_ids_type db_ids;

		for (std::size_t size = 0; size < _messages.size(); size++) {
			if (!to_json(messages, message, db_ids)) {
				return false;
			}

			if (sizeof(HeaderSend) + message.size() <
			    m_config.gwmpLimitDatagram) {
				return sendUplink(message, db_ids);

			} else {
				messages.pop_back();
			}
		}

		logr::error << "Failed to serialize message: "
			    << "Datagram limit reached" << logr::endl;
		return false;

	} catch (const std::exception &e) {
		logr::error << "Failed to serialize message: " << e.what()
			    << logr::endl;
		return false;
	}
}

bool GWMP::acknowledger(void)
{
	std::string message;

	if (!m_uplinkSocket.recv(message)) {
		if (!running()) {
			return false;
		}

		logr::error << "Failed to receive uplink acknowledge"
			    << logr::endl;
		return true;
	}

	HeaderRecv header = to_header(message);

	if (header.type == MessageType::PUSH_ACK) {
		m_uplinkManager.acknowledged(header.id);
		m_statistics.uplink_acknowledged++;
	}

	return true;
}

bool GWMP::heartbeat(void)
{
	HeaderSend header = {
		.version = Version::V2,
		.id = m_downlinkManager.next_id(),
		.type = MessageType::PULL_DATA,
		.gateway_id = m_config.gatewayId,
	};

	if (!m_downlinkSocket.send(to_string(header))) {
		if (!running()) {
			return false;
		}

		logr::error << "Failed to send heartbeat" << logr::endl;

	} else {
		m_downlinkManager.sent(header.id);
		m_statistics.heartbeat_sent++;
	}

	if (!sleep_for(m_config.gwmpPeriodHeartbeat)) {
		return false;
	}

	return true;
}

#define diff_statistic(diff, current, old, name)                               \
	diff.name = current.name - old.name;

#define diff_statistics(diff, current, old)                                    \
	diff_statistic(diff, current, old, uplink_forwarded);                  \
	diff_statistic(diff, current, old, uplink_sent);                       \
	diff_statistic(diff, current, old, uplink_acknowledged);               \
	diff_statistic(diff, current, old, heartbeat_sent);                    \
	diff_statistic(diff, current, old, heartbeat_acknowledged);            \
	diff_statistic(diff, current, old, downlink_received);                 \
	diff_statistic(diff, current, old, downlink_acknowledged);             \
	old = current;

#define diff_statistic_lorad(diff, current, old, name)                         \
	if (old.has_##name && current.has_##name) {                            \
		diff.has_##name = true;                                        \
		diff.name = current.name - old.name;                           \
	} else if (!old.has_##name && current.has_##name) {                    \
		diff.has_##name = true;                                        \
		diff.name = current.name;                                      \
	} else {                                                               \
		diff.has_##name = false;                                       \
	}

#define diff_statistics_lorad(diff, current, old)                              \
	diff_statistic_lorad(diff, current, old, uplink_received);             \
	diff_statistic_lorad(diff, current, old, uplink_invalid);              \
	diff_statistic_lorad(diff, current, old, uplink_valid);                \
	diff_statistic_lorad(diff, current, old, uplink_bytes_received);       \
	diff_statistic_lorad(diff, current, old, downlink_requested);          \
	diff_statistic_lorad(diff, current, old, downlink_scheduled);          \
	diff_statistic_lorad(diff, current, old, downlink_transmitted);        \
	diff_statistic_lorad(diff, current, old, downlink_bytes_transmitted);  \
	diff_statistic_lorad(diff, current, old, gps_lost_pulse);              \
	old = current;

bool GWMP::statistics(void)
{
	static std::time_t boot = std::time(nullptr);

	if (!sleep_for(m_config.gwmpPeriodStatistics)) {
		return false;
	}

	static LoradStatistics old_lorad = LoradStatistics{};
	static Statistics old_statistics = Statistics{};
	LoradStatistics current_lorad{}, lorad{};
	Statistics statistics{};

	diff_statistics(statistics, m_statistics, old_statistics);

	try {
		json root, stat;

		stat["time"] = message::time::to_string(
			std::time(nullptr),
			message::time::StringFormat::GWMP_EXPANDED);
		stat["boot"] = message::time::to_string(
			boot, message::time::StringFormat::GWMP_EXPANDED);
		stat["rxfw"] = statistics.uplink_forwarded;
		stat["ackr"] =
			setprecision(static_cast<json::number_float_t>(
					     statistics.uplink_acknowledged) *
					     100 / statistics.uplink_sent,
				     1);
		stat["dwnb"] = statistics.downlink_received;
		stat["ping"] = static_cast<int>(m_downlinkManager.ping());

		if (m_api.requestStatistics(current_lorad)) {
			diff_statistics_lorad(lorad, current_lorad, old_lorad);

			if (lorad.has_uplink_received) {
				stat["rxnb"] = lorad.uplink_received;
			}

			if (lorad.has_uplink_valid) {
				stat["rxok"] = lorad.uplink_valid;
			}

			if (lorad.has_downlink_transmitted) {
				stat["txnb"] = lorad.downlink_transmitted;
			}

			if (current_lorad.has_gps_latitude) {
				stat["lati"] = current_lorad.gps_latitude;
			}

			if (current_lorad.has_gps_longitude) {
				stat["long"] = current_lorad.gps_longitude;
			}

			if (current_lorad.has_gps_altitude) {
				stat["alti"] = current_lorad.gps_altitude;
			}

			if (lorad.has_gps_lost_pulse) {
				stat["lpps"] = lorad.gps_lost_pulse;
			}

			if (current_lorad.has_radio_temperature) {
				stat["temp"] = static_cast<int>(
					current_lorad.radio_temperature);
			}

			if (current_lorad.has_version_fpga) {
				stat["fpga"] = current_lorad.version_fpga;
			}

			if (current_lorad.has_version_dsp) {
				stat["dsp"] = current_lorad.version_dsp;
			}

			if (current_lorad.has_version_hal) {
				stat["hal"] = current_lorad.version_hal;
			}
		}

		root["stat"] = stat;

		sendUplink(root.dump());

	} catch (const std::exception &e) {
		logr::error << "Failed to serialize message: " << e.what()
			    << logr::endl;
		return false;
	}

	return true;
}

bool GWMP::timeout(void)
{
	if (!sleep_for(std::chrono::seconds(1))) {
		return false;
	}

	m_downlinkManager.clear();
	m_uplinkManager.clear();

	return true;
}
