// Copyright (c) 2018 Kerlink SA. All rights reserved.

#ifndef DEPRECATED_FEATURES
#error "Deprecated features must be enabled to use this header"
#endif

#ifndef _GWMP_H_
#define _GWMP_H_

#include <condition_variable>
#include <cstdint>
#include <list>
#include <map>
#include <mutex>
#include <string>
#include <vector>

extern "C" {
#include <sys/socket.h>
}

#include "utils/service.h"

#include "api.h"
#include "config.h"
#include "database.h"
#include "message/container.h"
#include "statistics.h"

namespace udp {
class Socket
{
public:
	Socket(void);
	~Socket(void);

	bool create(const char *_node, const char *_service);
	bool create(const std::string &_node, const std::string &_service);
	void shutdown(void);
	bool reset(void);
	bool running(void) const;

	bool send(const std::string &_message);
	bool recv(std::string &_message);

private:
	std::string m_node;
	std::string m_service;

	int m_sockfd;

	std::condition_variable m_condition;
	std::mutex m_mutex;
	bool m_resetting;

	void close(void);
	void wait_for_reset(void);

	sockaddr_storage m_sockaddr;
	socklen_t m_sockaddrlen;
};
}

class GWMP : public service::Service
{
public:
	using id_type = std::uint16_t;
	using message_id_type = message::uint_type;
	using db_ids_type = Database::ids_type;
	using handler_type = std::function<void(void)>;
	using timeout_type = std::chrono::seconds;

	static constexpr timeout_type TIMEOUT = timeout_type(3);

	GWMP(const Config &_config,
	     API &_api,
	     Database &_database,
	     Statistics &_statistics);
	~GWMP(void);

	virtual bool start(void);
	virtual void shutdown(void);
	virtual bool running(void) const;

	void registerHandler(handler_type _acknowledgedHandler,
			     handler_type _timedOutHandler);

	using event_type = message::event::Transmission;
	using downlink_type = message::downlink::Container;
	using uplink_type = message::uplink::Container;
	using uplinks_type = std::vector<uplink_type>;

	bool sendEvent(const event_type &_message);
	bool recvDownlink(downlink_type &_message);
	bool sendUplink(const uplink_type &_message);
	bool sendUplink(const uplinks_type &_messages);

private:
	const Config &m_config;
	API &m_api;
	Database &m_database;
	Statistics &m_statistics;

	udp::Socket m_downlinkSocket;
	udp::Socket m_uplinkSocket;

	class Manager
	{
	public:
		using ping_type = float;

		Manager(const std::string &_name);

		void attachDatabase(Database *_database);
		void registerHandler(handler_type _maxTimedOutHandler,
				     std::size_t _maxTimedOut);
		void registerHandler(handler_type _acknowledgedHandler,
				     handler_type _timedOutHandler);

		id_type next_id(void);

		void sent(id_type _id, const db_ids_type &_db_ids = {});
		void acknowledged(id_type _id);
		void clear(void);

		ping_type ping(void) const;

	private:
		using clock_type = std::chrono::steady_clock;
		using time_type = std::chrono::time_point<clock_type>;

		struct history_value_type
		{
			time_type time;
			db_ids_type db_ids;
		};

		using duration_type =
			std::chrono::duration<ping_type, std::milli>;

		void add_ping(const ping_type &_ping);
		void add_ping(const duration_type &_duration);

		std::string m_name;

		id_type m_nextId;
		std::map<id_type, history_value_type> m_history;
		std::mutex m_mutex;

		std::size_t m_maxPings;
		std::list<ping_type> m_pings;

		Database *m_database;

		std::size_t m_timedOut;
		std::size_t m_maxTimedOut;
		handler_type m_maxTimedOutHandler;

		handler_type m_acknowledgedHandler;
		handler_type m_timedOutHandler;
	};

	Manager m_downlinkManager;
	Manager m_uplinkManager;

	void restart(void);

	bool sendUplink(const std::string &_message,
			const db_ids_type &_db_ids = {});

	bool acknowledger(void);
	bool heartbeat(void);
	bool statistics(void);
	bool timeout(void);
};

#endif
