// Copyright (c) 2018 Kerlink SA. All rights reserved.

#include <cstdlib>
#include <exception>
#include <iostream>
#include <string>
#include <vector>

#include "third-party/cxxopts.h"

#include "utils/daemon.h"
#include "utils/logger.h"
#include "utils/service.h"

#include "api.h"
#include "config.h"
#include "database.h"
#include "filter.h"
#include "forwarder.h"
#include "gwmp.h"
#include "statistics.h"

#ifndef RUNSTATEDIR
#define RUNSTATEDIR "/var/run/lora"
#endif

namespace {
std::vector<service::Service *> g_services;

void signalHandler(int _signal)
{
	switch (_signal) {
	case SIGINT:
	case SIGQUIT:
	case SIGTERM:
		logr::notice << "Stopping lorafwd" << logr::endl;

		for (auto &service : g_services) {
			service->shutdown();
		}

		break;
	}
}

int realmain(int _argc, char **_argv)
{
	logr::level verbosity = logr::level::warning;
	std::vector<std::string> files;
	bool foreground = false;

	cxxopts::Options options("lorafwd", "The LoRa forwarder");

	// clang-format off
	options.add_options()
		("c, config", "Use the given configuration file.",
		 cxxopts::value(files), "FILE")
		("f, foreground", "Do not run in background.",
		 cxxopts::value(foreground))
		("v, verbose", "Turn on verbose output. "
			       "This option can be used multiple times.")
		("version", "Print version string and exit.")
		("h, help", "Print this help message and exit.");
	// clang-format on

	options.positional_help("FILE...").parse_positional("config");

	try {
		auto opts = options.parse(_argc, _argv);

		if (opts.count("help")) {
			std::cout << options.help();
			return EXIT_SUCCESS;
		}

		if (opts.count("version")) {
			std::cout << "lorafwd " << VERSION << std::endl;
			return EXIT_SUCCESS;
		}

		verbosity += opts.count("verbose");

	} catch (const cxxopts::OptionParseException &e) {
		std::cerr << e.what() << std::endl;
		std::cerr << options.help();
		return EXIT_FAILURE;
	}

	logr::setLevel(verbosity);

	// Parse all configuration files before to be daemonized. So the stderr
	// is still available and parsing errors can be printed directly on it.
	Config config;

	for (const auto &file : files) {
		logr::info << "Parsing configuration file: " << file
			   << logr::endl;

		if (!config.parse_file(file)) {
			logr::critical << "Failed to parse configuration file: "
				       << file << logr::endl;
			return EXIT_FAILURE;
		}
	}

	Filter filter(config);
	Statistics statistics{};

	API api(config, statistics);
	Database database(config);
	GWMP gwmp(config, api, database, statistics);
	Forwarder forwarder(config, api, filter, database, gwmp);

	g_services.push_back(&api);
	g_services.push_back(&database);
	g_services.push_back(&gwmp);
	g_services.push_back(&forwarder);

	// Run the process in background as soon as possible.
	//
	// As the process will fork to be turned into a daemon, the objects
	// which needs to be cleanly deleted in the parent must be created
	// before this point.
	daemon::PidFile pidfile;
	int pid = 0;

	if (foreground == false && (pid = daemon::daemonize()) < 0) {
		logr::critical << "Failed to daemonize" << logr::endl;
		return EXIT_FAILURE;
	}

	if (pid > 0) {
		return EXIT_SUCCESS;
	}

	if (!pidfile.create(RUNSTATEDIR "/lorafwd.pid")) {
		logr::critical << "Failed to create PID file" << logr::endl;
		return EXIT_FAILURE;
	}

	if (!daemon::handleSignals({{SIGINT, signalHandler},
				    {SIGQUIT, signalHandler},
				    {SIGTERM, signalHandler}})) {
		logr::critical << "Failed to handle signals" << logr::endl;
		return EXIT_FAILURE;
	}

	logr::notice << "Starting lorafwd " << VERSION << logr::endl;

	// Dump the configuration.
	logr::info << config << logr::endl;

	// It is time to start the services.
	for (auto &service : g_services) {
		if (!service->start()) {
			logr::critical << "Failed to start service"
				       << logr::endl;
			return EXIT_FAILURE;
		}
	}

	// The main process will then wait until the end of all his children.
	for (auto &service : g_services) {
		if (!service->join()) {
			logr::critical << "Failed to join service"
				       << logr::endl;
			return EXIT_FAILURE;
		}
	}

	return EXIT_SUCCESS;
}
}

// The guidelines for this project is to NOT use the exceptions. But
// unfortunately some standard objects and included libraries are using it. The
// exceptions which can occurred should be caught and handled as soon as
// possible.
//
// The main() function is only an exception wrapper for the realmain() in order
// to log the remaining exceptions which are not handled.
int main(int _argc, char **_argv)
{
	int status = EXIT_FAILURE;

	logr::open("lorafwd",
		   logr::level::warning,
		   LOG_PID | LOG_PERROR,
		   LOG_LOCAL1);

	try {
		status = realmain(_argc, _argv);

	} catch (const std::exception &e) {
		logr::alert << "Uncaught exception: " << e.what() << logr::endl;

	} catch (...) {
		logr::alert << "Uncaught exception" << logr::endl;
	}

	return status;
}
