// Copyright (c) 2018 Kerlink SA. All rights reserved.

#include "message/data.h"

#include "utils/lorawan.h"
#include "utils/stream.h"

#define PREFIX std::endl << "| "

namespace message {
namespace radio_id {
uint_type to_channel(uint_type _radio_id)
{
	return ((_radio_id & 0x0000000F) >> 0);
}

uint_type to_demodulator(uint_type _radio_id)
{
	return ((_radio_id & 0x000000F0) >> 4);
}

uint_type to_transceiver(uint_type _radio_id)
{
	return ((_radio_id & 0x00000F00) >> 8);
}

uint_type to_board(uint_type _radio_id)
{
	return ((_radio_id & 0x0000F000) >> 12);
}

uint_type to_antenna(uint_type _radio_id)
{
	return ((_radio_id & 0x000F0000) >> 16);
}

uint_type to_zone(uint_type _radio_id)
{
	return ((_radio_id & 0x00F00000) >> 20);
}

uint_type to_radio_id(uint_type _channel,
		      uint_type _demodulator,
		      uint_type _transceiver,
		      uint_type _board,
		      uint_type _antenna,
		      uint_type _zone)
{
	return ((_channel & 0x0F) << 0) | ((_demodulator & 0x0F) << 4) |
	       ((_transceiver & 0x0F) << 8) | ((_board & 0x0F) << 12) |
	       ((_antenna & 0x0F) << 16) | ((_zone & 0x0F) << 20);
}
}

namespace downlink {
namespace {
struct TimingVisitor
{
	std::ostream &m_os;

	TimingVisitor(std::ostream &_os) : m_os{_os} {}

	void operator()(uint_type _counter)
	{
		m_os << ", on counter " << _counter;
	}

	void operator()(const message::time::AccurateContinuousTime &_time)
	{
		switch (_time.format) {
		case message::time::ContinuousTimeFormat::TAI_TIME:
			m_os << ", on TAI time ";
			break;
		case message::time::ContinuousTimeFormat::GPS_TIME:
			m_os << ", on GPS time ";
			break;
		}

		m_os << message::time::to_string(_time);
	}

	void operator()(const TimeBoundaries &_boundaries)
	{
		if (_boundaries.has_from_counter ||
		    _boundaries.has_to_counter) {
			m_os << ",";

			if (_boundaries.has_from_counter) {
				m_os << " from counter "
				     << _boundaries.from_counter;
			}

			if (_boundaries.has_to_counter) {
				m_os << " to counter "
				     << _boundaries.to_counter;
			}

		} else {
			m_os << ", as soon as possible";
		}
	}
};

std::ostream &operator<<(std::ostream &_os, const Timing &_message)
{
	_os << "transmission (" << utils::to_hex(_message.transmission_id)
	    << ")";

	_os << ", priority " << _message.priority;

	mpark::visit(TimingVisitor(_os), _message.timing);

	if (_message.pre_delay || _message.post_delay) {
		_os << ", delay";

		if (_message.pre_delay) {
			_os << " -" << _message.pre_delay << " usec";
		}

		if (_message.post_delay) {
			_os << " +" << _message.pre_delay << " usec";
		}
	}

	return _os;
}
}
}

namespace uplink {
namespace {
std::string to_string(CRCStatus _crc)
{
	switch (_crc) {
	case CRCStatus::CRC_OK:
		return "ok";
	case CRCStatus::CRC_ERROR:
		return "error";
	case CRCStatus::CRC_NONE:
		return "none";
	default:
		return "unknown";
	}
}

std::ostream &operator<<(std::ostream &_os,
			 const AbstractRadioDemodulator &_message)
{
	_os << "demodulator counter " << _message.counter;

	if (_message.has_time) {
		switch (_message.time.format) {
		case message::time::TimeFormat::UNIX_TIME:
			_os << ", UTC time ";
			break;
		case message::time::TimeFormat::TAI_TIME:
			_os << ", TAI time ";
			break;
		case message::time::TimeFormat::GPS_TIME:
			_os << ", GPS time ";
			break;
		}

		_os << message::time::to_string(_message.time);
	}

	if (_message.has_rssi) {
		_os << ", rssi " << _message.rssi << " dB";
	}

	return _os;
}
}
}

namespace lora {
namespace {
std::string to_string(Bandwidth _bandwidth)
{
	switch (_bandwidth) {
	case Bandwidth::BW_125K:
		return "125 kHz";
	case Bandwidth::BW_250K:
		return "250 kHz";
	case Bandwidth::BW_500K:
		return "500 kHz";
	default:
		return "unknown";
	}
}

std::string to_string(SpreadingFactor _spreading_factor)
{
	switch (_spreading_factor) {
	case SpreadingFactor::SF_7:
		return "7";
	case SpreadingFactor::SF_8:
		return "8";
	case SpreadingFactor::SF_9:
		return "9";
	case SpreadingFactor::SF_10:
		return "10";
	case SpreadingFactor::SF_11:
		return "11";
	case SpreadingFactor::SF_12:
		return "12";
	default:
		return "unknown";
	}
}

std::string to_string(CodingRate _coding_rate)
{
	switch (_coding_rate) {
	case CodingRate::CR_4_5:
		return "4/5";
	case CodingRate::CR_4_6:
		return "4/6";
	case CodingRate::CR_4_7:
		return "4/7";
	case CodingRate::CR_4_8:
		return "4/8";
	default:
		return "unknown";
	}
}

std::ostream &operator<<(std::ostream &_os, const Configuration &_message)
{
	_os << "radio (" << utils::to_hex(_message.radio_id) << ")";
	_os << ", channel " << _message.frequency / 1e6f << " MHz";
	_os << ", bw " << to_string(_message.bandwidth);
	_os << ", sf " << to_string(_message.spreading_factor);
	_os << ", cr " << to_string(_message.coding_rate);

	if (_message.has_power) {
		_os << ", power " << _message.power << " dB";
	}

	for (const auto &timing : _message.timings) {
		_os << PREFIX << "  - ";
		downlink::operator<<(_os, timing);
	}

	return _os;
}

std::ostream &operator<<(std::ostream &_os, const RadioDemodulator &_message)
{
	uplink::operator<<(
		_os, static_cast<uplink::AbstractRadioDemodulator>(_message));

	if (_message.has_snr) {
		auto &snr = _message.snr;

		_os << ", snr ";

		if (snr.has_min) {
			_os << snr.min << "< ";
		}

		_os << snr.snr;

		if (snr.has_max) {
			_os << " <" << snr.max;
		}

		_os << " dB";
	}

	return _os;
}

struct LocalizationVisitor
{
	std::ostream &m_os;

	LocalizationVisitor(std::ostream &_os) : m_os{_os} {}

	void operator()(const std::string &_encrypted)
	{
		m_os << " encrypted " << utils::to_hex(_encrypted);
	}

	void operator()(uint_type _decrypted)
	{
		m_os << " decrypted " << _decrypted << " ns";
	}
};

std::ostream &operator<<(std::ostream &_os, const RadioLocalization &_message)
{
	_os << "localization";

	if (_message.has_nanoseconds) {
		mpark::visit(LocalizationVisitor(_os), _message.nanoseconds);

	} else {
		_os << " unknown";
	}

	if (_message.has_rssi) {
		_os << ", rssi " << _message.rssi << " dB";
	}

	if (_message.has_standard_deviation && _message.standard_deviation) {
		_os << ", standard deviation " << _message.standard_deviation
		    << " dB";
	}

	if (_message.has_frequency_offset && _message.frequency_offset) {
		_os << ", frequency offset " << _message.frequency_offset
		    << " Hz";
	}

	if (_message.has_status && _message.status) {
		_os << ", status (" << utils::to_hex(_message.status) << ")";
	}

	if (_message.has_version) {
		_os << ", version " << _message.version;
	}

	if (_message.has_delta_alternate && _message.delta_alternate) {
		_os << ", delta alternate " << _message.delta_alternate
		    << " ns";
	}

	return _os;
}

std::ostream &operator<<(std::ostream &_os, const Radio &_message)
{
	_os << "radio (" << utils::to_hex(_message.radio_id) << ")";

	if (_message.has_demodulator) {
		_os << PREFIX << "  - " << _message.demodulator;
	}

	if (_message.has_localization) {
		_os << PREFIX << "  - " << _message.localization;
	}

	return _os;
}
}

std::ostream &operator<<(std::ostream &_os, const Downlink &_message)
{
	_os << PREFIX;
	_os << "lora downlink (" << utils::to_hex(_message.message_id) << ")";
	_os << ", payload " << _message.payload.size() << " B";
	_os << ", required " << _message.required_transmissions;
	_os << ", preamble " << _message.preamble << " B";
	_os << ", header "
	    << ((_message.header_disabled) ? "disabled" : "enabled");
	_os << ", crc " << ((_message.crc_enabled) ? "enabled" : "disabled");
	_os << ", polarity "
	    << ((_message.polarity_inverted) ? "inverted" : "normal");

	_os << PREFIX << utils::LoRaWAN(_message.payload);

	for (const auto &configuration : _message.configurations) {
		_os << PREFIX << " - " << configuration;
	}

	return _os;
}

std::ostream &operator<<(std::ostream &_os, const Uplink &_message)
{
	_os << PREFIX;
	_os << "lora uplink (" << utils::to_hex(_message.message_id) << ")";

	if (_message.has_data) {
		auto &data = _message.data;

		_os << ", payload " << data.payload.size() << " B";

		if (data.has_frequency) {
			_os << ", channel " << data.frequency / 1e6f << " MHz";
		}

		_os << ", crc " << uplink::to_string(data.crc);
		_os << ", bw " << to_string(data.bandwidth);
		_os << ", sf " << to_string(data.spreading_factor);
		_os << ", cr " << to_string(data.coding_rate);

		if (data.crc != message::uplink::CRCStatus::CRC_ERROR) {
			_os << PREFIX << utils::LoRaWAN(data.payload);
		}
	}

	for (const auto &radio : _message.radios) {
		_os << PREFIX << " - " << radio;
	}

	return _os;
}
}

namespace gfsk {
namespace {
std::string to_string(CRCType _crc)
{
	switch (_crc) {
	case CRCType::CRC_CCITT:
		return "CCITT";
	case CRCType::CRC_IBM:
		return "IBM";
	case CRCType::CRC_DISABLED:
		return "disabled";
	default:
		return "unknown";
	}
}

std::string to_string(EncodingType _encoding)
{
	switch (_encoding) {
	case EncodingType::ENC_WHITENING:
		return "whitening";
	case EncodingType::ENC_MANCHESTER:
		return "manchester";
	case EncodingType::ENC_DISABLED:
		return "disabled";
	default:
		return "unknown";
	}
}

std::string to_string(Bandwidth _bandwidth)
{
	switch (_bandwidth) {
	case Bandwidth::BW_7K8:
		return "7.8125 kHz";
	case Bandwidth::BW_15K6:
		return "15.625 kHz";
	case Bandwidth::BW_31K2:
		return "31.25 kHz";
	case Bandwidth::BW_62K5:
		return "62.5 kHz";
	case Bandwidth::BW_125K:
		return "125 kHz";
	case Bandwidth::BW_250K:
		return "250 kHz";
	case Bandwidth::BW_500K:
		return "500 kHz";
	default:
		return "unknown";
	}
}

std::ostream &operator<<(std::ostream &_os, const Configuration &_message)
{
	_os << "radio (" << utils::to_hex(_message.radio_id) << ")";
	_os << ", channel " << _message.frequency / 1e6f << " MHz";
	_os << ", br " << _message.bitrate / 1e3f << " kbps";
	_os << ", fdev " << _message.frequency_deviation / 1e3f << " kHz";

	if (_message.has_power) {
		_os << ", power " << _message.power << " dB";
	}

	for (const auto &timing : _message.timings) {
		_os << PREFIX << "  - ";
		downlink::operator<<(_os, timing);
	}

	return _os;
}

std::ostream &operator<<(std::ostream &_os, const Radio &_message)
{
	_os << "radio (" << utils::to_hex(_message.radio_id) << ")";

	if (_message.has_demodulator) {
		_os << PREFIX << "  - ";
		uplink::operator<<(
			_os,
			static_cast<uplink::AbstractRadioDemodulator>(
				_message.demodulator));
	}

	return _os;
}
}

std::ostream &operator<<(std::ostream &_os, const Downlink &_message)
{
	_os << PREFIX;
	_os << "gfsk downlink (" << utils::to_hex(_message.message_id) << ")";
	_os << ", payload " << _message.payload.size() << " B";
	_os << ", required " << _message.required_transmissions;
	_os << ", preamble " << _message.preamble << " B";
	_os << ", length " << ((_message.length_fixed) ? "fixed" : "variable");
	_os << ", crc " << to_string(_message.crc);
	_os << ", encoding " << to_string(_message.encoding);

	_os << PREFIX << utils::LoRaWAN(_message.payload);

	for (const auto &configuration : _message.configurations) {
		_os << PREFIX << " - " << configuration;
	}

	return _os;
}

std::ostream &operator<<(std::ostream &_os, const Uplink &_message)
{
	_os << PREFIX;
	_os << "gfsk uplink (" << utils::to_hex(_message.message_id) << ")";

	if (_message.has_data) {
		auto &data = _message.data;

		_os << ", payload " << data.payload.size() << " B";

		if (data.has_frequency) {
			_os << ", channel " << data.frequency / 1e6f << " MHz";
		}

		_os << ", crc " << uplink::to_string(data.crc);
		_os << ", bw " << to_string(data.bandwidth);
		_os << ", br " << data.bitrate / 1e3f << " kbps";

		if (data.crc != message::uplink::CRCStatus::CRC_ERROR) {
			_os << PREFIX << utils::LoRaWAN(data.payload);
		}
	}

	for (const auto &radio : _message.radios) {
		_os << PREFIX << " - " << radio;
	}

	return _os;
}
}
}
