// Copyright (c) 2018 Kerlink SA. All rights reserved.

#ifndef _MESSAGE_DATA_H_
#define _MESSAGE_DATA_H_

#include <chrono>
#include <cstdint>
#include <ostream>
#include <string>
#include <vector>

#include "third-party/variant.h"

#include "message/common.h"

namespace message {
using clock_type = std::chrono::steady_clock;
using recv_time_type = std::chrono::milliseconds;

static constexpr recv_time_type RX1 = recv_time_type(1000);
static constexpr recv_time_type RX2 = recv_time_type(2000);

namespace radio_id {
uint_type to_channel(uint_type _radio_id);
uint_type to_demodulator(uint_type _radio_id);
uint_type to_transceiver(uint_type _radio_id);
uint_type to_board(uint_type _radio_id);
uint_type to_antenna(uint_type _radio_id);
uint_type to_zone(uint_type _radio_id);
uint_type to_radio_id(uint_type _channel,
		      uint_type _demodulator,
		      uint_type _transceiver,
		      uint_type _board,
		      uint_type _antenna,
		      uint_type _zone);
}

struct AbstractMessage
{
	virtual ~AbstractMessage(void) {}
	uint_type message_id;
};

namespace downlink {
struct TimeBoundaries
{
	bool_type has_from_counter;
	uint_type from_counter;
	bool_type has_to_counter;
	uint_type to_counter;
};

struct Timing
{
	uint_type transmission_id;
	uint_type priority;
	mpark::variant<uint_type, time::AccurateContinuousTime, TimeBoundaries>
		timing;
	uint_type post_delay;
	uint_type pre_delay;
};

struct AbstractConfiguration
{
	virtual ~AbstractConfiguration(void) {}
	uint_type radio_id;
	uint_type frequency;
	bool_type has_power;
	float_type power;
	std::vector<Timing> timings;
};

struct AbstractMessage : virtual message::AbstractMessage
{
	virtual ~AbstractMessage(void) {}
	std::string payload;
	uint_type required_transmissions;
	uint_type preamble;
};
}

namespace uplink {
enum class CRCStatus : uint_type {
	CRC_OK = 0,
	CRC_ERROR = 1,
	CRC_NONE = 2,
};

struct AbstractData
{
	virtual ~AbstractData(void) {}
	recv_time_type recv_time;
	std::string payload;
	bool_type has_frequency;
	uint_type frequency;
	CRCStatus crc;
};

struct AbstractRadioDemodulator
{
	virtual ~AbstractRadioDemodulator(void) {}
	recv_time_type recv_time;
	uint_type counter;
	bool_type has_time;
	time::Time time;
	bool_type has_rssi;
	float_type rssi;
};

struct AbstractRadio
{
	virtual ~AbstractRadio(void) {}
	uint_type radio_id;
	bool_type has_demodulator;
};

struct AbstractMessage : virtual message::AbstractMessage
{
	virtual ~AbstractMessage(void) {}
	bool_type has_data;
};
}

namespace lora {
enum class Bandwidth : uint_type {
	BW_125K = 0,
	BW_250K = 1,
	BW_500K = 2,
};

enum class CodingRate : uint_type {
	CR_1 = 0,
	CR_4_5 = 0,
	CR_2 = 1,
	CR_4_6 = 1,
	CR_2_3 = 1,
	CR_3 = 2,
	CR_4_7 = 2,
	CR_4 = 3,
	CR_4_8 = 3,
	CR_2_4 = 3,
	CR_1_2 = 3,
};

enum class SpreadingFactor : uint_type {
	SF_7 = 0,
	SF_8 = 1,
	SF_9 = 2,
	SF_10 = 3,
	SF_11 = 4,
	SF_12 = 5,
};

struct Configuration : virtual downlink::AbstractConfiguration
{
	Bandwidth bandwidth;
	CodingRate coding_rate;
	SpreadingFactor spreading_factor;
};

struct Downlink : virtual downlink::AbstractMessage
{
	bool_type crc_enabled;
	bool_type header_disabled;
	bool_type polarity_inverted;
	std::vector<Configuration> configurations;
};

struct Data : virtual uplink::AbstractData
{
	Bandwidth bandwidth;
	CodingRate coding_rate;
	SpreadingFactor spreading_factor;
};

struct SNR
{
	float_type snr;
	bool_type has_max;
	float_type max;
	bool_type has_min;
	float_type min;
};

struct RadioDemodulator : virtual uplink::AbstractRadioDemodulator
{
	bool_type has_snr;
	SNR snr;
};

struct RadioLocalization
{
	recv_time_type recv_time;
	bool_type has_nanoseconds;
	mpark::variant<std::string, uint_type> nanoseconds;
	bool_type has_rssi;
	float_type rssi;
	bool_type has_standard_deviation;
	uint_type standard_deviation;
	bool_type has_frequency_offset;
	int_type frequency_offset;
	bool_type has_status;
	uint_type status;
	bool_type has_version;
	uint_type version;
	bool_type has_delta_alternate;
	int_type delta_alternate;
};

struct Radio : virtual uplink::AbstractRadio
{
	RadioDemodulator demodulator;
	bool_type has_localization;
	RadioLocalization localization;
};

struct Uplink : virtual uplink::AbstractMessage
{
	Data data;
	std::vector<Radio> radios;
};

std::ostream &operator<<(std::ostream &_os, const Downlink &_message);
std::ostream &operator<<(std::ostream &_os, const Uplink &_message);
}

namespace gfsk {
struct Configuration : virtual downlink::AbstractConfiguration
{
	uint_type bitrate;
	uint_type frequency_deviation;
};

enum class CRCType : uint_type {
	CRC_CCITT = 0,
	CRC_IBM = 1,
	CRC_DISABLED = 2,
};

enum class EncodingType : uint_type {
	ENC_WHITENING = 0,
	ENC_MANCHESTER = 1,
	ENC_DISABLED = 2,
};

struct Downlink : virtual downlink::AbstractMessage
{
	bool_type length_fixed;
	CRCType crc;
	EncodingType encoding;
	std::vector<Configuration> configurations;
};

enum class Bandwidth : uint_type {
	BW_125K = 0,
	BW_250K = 1,
	BW_500K = 2,
	BW_7K8 = 3,
	BW_15K6 = 4,
	BW_31K2 = 5,
	BW_62K5 = 6,
};

struct Data : virtual uplink::AbstractData
{
	Bandwidth bandwidth;
	uint_type bitrate;
};

struct RadioDemodulator : virtual uplink::AbstractRadioDemodulator
{
};

struct Radio : virtual uplink::AbstractRadio
{
	RadioDemodulator demodulator;
};

struct Uplink : virtual uplink::AbstractMessage
{
	Data data;
	std::vector<Radio> radios;
};

std::ostream &operator<<(std::ostream &_os, const Downlink &_message);
std::ostream &operator<<(std::ostream &_os, const Uplink &_message);
}
}

#endif
