// Copyright (c) 2018 Kerlink SA. All rights reserved.

#include "utils/daemon.h"

#include <sstream>

namespace libc {
// This namespace ensure that the "daemon" namespace define below does not
// conflict with the BSD daemon(3) function.

extern "C" {
#include <fcntl.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
}
}

#include "utils/logger.h"

namespace daemon {

// The daemonize() function will follow the steps define here:
// https://www.freedesktop.org/software/systemd/man/daemon.html
//
// This function should be called at the very beginning of the process, this way
// the steps 1 to 4 can be skipped.
//
// This function handle the steps 5 to 11. The step 12 is handled by the PidFile
// class.
//
// The step 13 is not applicable. And to make it easier the steps 14 and 15 are
// not implemented.
int daemonize(void)
{
	pid_t pid;
	int fd;

	if ((pid = libc::fork()) < 0) {
		logr::error << "Failed to create a child process"
			    << logr::eendl;
		return -1;
	}

	if (pid > 0) {
		return pid;
	}

	if (libc::setsid() < 0) {
		logr::error << "Failed to create a new session" << logr::eendl;
		return -1;
	}

	if ((pid = libc::fork()) < 0) {
		logr::error << "Failed to create a second child process"
			    << logr::eendl;
		return -1;
	}

	if (pid > 0) {
		return pid;
	}

	if (libc::close(STDIN_FILENO) < 0) {
		logr::error << "Failed to close stdin" << logr::eendl;
		return -1;
	}

	if ((fd = libc::open("/dev/null", O_RDONLY)) < 0) {
		logr::error << "Failed to open stdin" << logr::eendl;
		return -1;
	}

	if (fd != STDIN_FILENO) {
		logr::error << "Mismatch stdin file descriptor" << logr::eendl;
		return -1;
	}

	if (libc::close(STDOUT_FILENO) < 0) {
		logr::error << "Failed to close stdout" << logr::eendl;
		return -1;
	}

	if ((fd = libc::open("/dev/null", O_WRONLY)) < 0) {
		logr::error << "Failed to open stdout" << logr::eendl;
		return -1;
	}

	if (fd != STDOUT_FILENO) {
		logr::error << "Mismatch stdout file descriptor" << logr::eendl;
		return -1;
	}

	if (libc::close(STDERR_FILENO) < 0) {
		logr::error << "Failed to close stderr" << logr::eendl;
		return -1;
	}

	if ((fd = libc::dup(STDOUT_FILENO)) < 0) {
		logr::error << "Failed to duplicate stderr" << logr::eendl;
		return -1;
	}

	if (fd != STDERR_FILENO) {
		logr::error << "Mismatch stderr file descriptor" << logr::eendl;
		return -1;
	}

	if (libc::chdir("/") < 0) {
		logr::error << "Failed to change working directory"
			    << logr::eendl;
		return -1;
	}

	libc::umask(0);

	return 0;
}

bool handleSignals(const std::vector<struct signal> &_signals)
{
	for (const auto &signal : _signals) {
		if (std::signal(signal.num, signal.handler) == SIG_ERR) {
			logr::error << "Failed to handle signal: "
				    << libc::strsignal(signal.num)
				    << logr::eendl;
			return false;
		}
	}

	return true;
}

PidFile::PidFile(void) : m_file(), m_fd(-1) {}

// To fit the RAII principle the pid file is removed when the object is deleted.
PidFile::~PidFile(void)
{
	if (m_fd >= 0 && libc::close(m_fd) < 0) {
		logr::warning << "Failed to close pid file: " << m_file
			      << logr::eendl;
	}

	if (m_file.empty() == false && std::remove(m_file.c_str()) < 0) {
		logr::warning << "Failed to delete pid file: " << m_file
			      << logr::eendl;
	}
}

// The lock on the pid file ensure that the process cannot be started more than
// once.
bool PidFile::create(const char *_file)
{
	std::stringstream pid;
	int fd;

	if ((fd = libc::open(_file, O_CREAT | O_WRONLY, 0644)) < 0) {
		logr::error << "Failed to open pid file: " << _file
			    << logr::eendl;
		return false;
	}

	if (libc::lockf(fd, F_TLOCK, 0) < 0) {
		logr::error << "Failed to lock pid file: " << _file
			    << logr::eendl;
		libc::close(fd);
		return false;
	}

	if (libc::ftruncate(fd, 0) < 0) {
		logr::error << "Failed to clear pid file: " << _file
			    << logr::eendl;
		libc::close(fd);
		return false;
	}

	pid << libc::getpid();

	if (libc::write(fd, pid.str().data(), pid.str().size()) < 0) {
		logr::error << "Failed to write pid file: " << _file
			    << logr::eendl;
		libc::close(fd);
		return false;
	}

	m_file = _file;
	m_fd = fd;

	return true;
}

bool PidFile::create(const std::string &_file) { return create(_file.c_str()); }
}
