// Copyright (c) 2018 Kerlink SA. All rights reserved.

#include "logger.h"

#include <cctype>
#include <cerrno>
#include <cstdint>
#include <cstring>
#include <iomanip>
#include <sstream>
#include <streambuf>

#include "stream.h"

namespace logr {
void open(const std::string &_identity,
	  level _level,
	  int _option,
	  int _facility)
{
	::openlog(_identity.c_str(), _option, _facility);
	setLevel(_level);
}

level &operator++(level &_level)
{
	switch (_level) {
	case level::emergency:
		return _level = level::alert;
	case level::alert:
		return _level = level::critical;
	case level::critical:
		return _level = level::error;
	case level::error:
		return _level = level::warning;
	case level::warning:
		return _level = level::notice;
	case level::notice:
		return _level = level::info;
	default:
		return _level = level::debug;
	}
}

level &operator+=(level &_level, std::size_t _i)
{
	for (std::size_t i = 0; i < _i; i++) {
		++_level;
	}

	return _level;
}

void setLevel(level _level)
{
	::setlogmask(LOG_UPTO(static_cast<int>(_level)));
}

// This Logger class, which inherit from std::streambuf, will be used to create
// the std::ostream objects needed to mimic the standard std::cout behavior. The
// sync() and overflow() methods are updated to use syslog(3) with the level
// given in the constructor.
class Logger : public std::streambuf
{
public:
	Logger(level _level) : m_priority(static_cast<int>(_level)) {}

	// In order to properly close the syslog file descriptor, closelog(3)
	// have to be used at the end of the process. The Logger destructor will
	// be called for each Logger objects (see below). But the closelog(3)
	// function can be used multiple times, so it is not a problem to call
	// it here. Moreover the user do not have to think about it later.
	~Logger(void) { ::closelog(); }

protected:
	int sync(void) override
	{
		if (m_buffer.empty()) {
			return 0;
		}

		::syslog(m_priority, "<%u> %s", m_priority, m_buffer.c_str());
		m_buffer.clear();

		return 0;
	}

	std::streambuf::int_type overflow(std::streambuf::int_type c) override
	{
		if (c == std::streambuf::traits_type::eof()) {
			sync();

		} else {
			m_buffer += static_cast<char>(c);
		}

		return c;
	}

private:
	int m_priority;
	std::string m_buffer;
};

Logger g_emergency(level::emergency);
Logger g_alert(level::alert);
Logger g_critical(level::critical);
Logger g_error(level::error);
Logger g_warning(level::warning);
Logger g_notice(level::notice);
Logger g_info(level::info);
Logger g_debug(level::debug);

std::ostream emergency(&g_emergency);
std::ostream alert(&g_alert);
std::ostream critical(&g_critical);
std::ostream error(&g_error);
std::ostream warning(&g_warning);
std::ostream notice(&g_notice);
std::ostream info(&g_info);
std::ostream debug(&g_debug);

std::ostream &endl(std::ostream &_os) { return std::flush(_os); }

std::ostream &eendl(std::ostream &_os)
{
	return std::flush(_os << ": " << std::strerror(errno));
}

Dump dump(const void *_data,
	  std::size_t _size,
	  const std::string &_prefix,
	  std::size_t _column)
{
	return {static_cast<const char *>(_data), _size, _prefix, _column};
}

Dump dump(const std::string &_data,
	  const std::string &_prefix,
	  std::size_t _column)
{
	return {_data.data(), _data.size(), _prefix, _column};
}

namespace {
std::ostream &dump(std::ostream &_os, const Dump &_dump)
{
	for (std::size_t i = 0; i < _dump.size;) {
		std::ostringstream hex, str;

		for (std::size_t j = 0; i < _dump.size && j < _dump.column;
		     i++, j++) {
			char d = _dump.data[i];
			hex << utils::to_hex(d) << " ";
			str << ((std::isprint(d)) ? d : '.');
		}

		_os << std::right << std::setw(10) << _dump.prefix << ":  "
		    << std::left << std::setw(3 * _dump.column) << hex.str()
		    << " |" << str.str() << "|" << endl;
	}

	return _os;
}
}

std::ostream &operator<<(std::ostream &_os, const Dump &_dump)
{
	return dump(_os, _dump);
}
}
