// Copyright (c) 2018 Kerlink SA. All rights reserved.

#ifndef _UTILS_LOGGER_H_
#define _UTILS_LOGGER_H_

#include <cstddef>
#include <ostream>
#include <string>

extern "C" {
#include <syslog.h>
}

// This logger module is design to mimic the std::cout behavior while using
// syslog(3). Global objects are used to start logging, from logr::emergency to
// logr::debug following the standard syslog(3) levels.
//
// Standard manipulators (e.g. std::hex) can be used directly on the stream. The
// logr::endl manipulator (which acts exactly like std::endl) have been defined
// for clarity only.
//
// The logr::eendl also append the standard error string (from errno) to the
// stream. This is useful when working with the standard C libraries.
//
// Dynamic log levels are also supported with the logr::setLevel() function.
//
// Last but not least, dump utilities functions have been added in order to
// easily display memory (to be used with network packets ...).

namespace logr {
enum class level : int {
	emergency = LOG_EMERG,
	alert = LOG_ALERT,
	critical = LOG_CRIT,
	error = LOG_ERR,
	warning = LOG_WARNING,
	notice = LOG_NOTICE,
	info = LOG_INFO,
	debug = LOG_DEBUG,
};

void open(const std::string &_identity,
	  level _level = level::warning,
	  int _option = LOG_PID | LOG_PERROR,
	  int _facility = LOG_DAEMON);

level &operator++(level &_level);
level &operator+=(level &_level, std::size_t _i);

void setLevel(level _level);

extern std::ostream emergency;
extern std::ostream alert;
extern std::ostream critical;
extern std::ostream error;
extern std::ostream warning;
extern std::ostream notice;
extern std::ostream info;
extern std::ostream debug;

std::ostream &endl(std::ostream &_os);
std::ostream &eendl(std::ostream &_os);

struct Dump
{
	const char *data;
	std::size_t size;
	const std::string &prefix;
	std::size_t column;
};

Dump dump(const void *_data,
	  std::size_t _size,
	  const std::string &_prefix = "dump",
	  std::size_t _column = 32);

Dump dump(const std::string &_data,
	  const std::string &_prefix = "dump",
	  std::size_t _column = 32);

std::ostream &operator<<(std::ostream &_os, const Dump &_dump);
}

#endif
