// Copyright (c) 2018 Kerlink SA. All rights reserved.

#include "lorawan.h"

#include "utils/logger.h"
#include "utils/stream.h"

namespace utils {
namespace {
extern "C" {
#include <endian.h>
}

std::uint16_t loratoh(std::uint16_t _value) { return le16toh(_value); }
std::uint32_t loratoh(std::uint32_t _value) { return le32toh(_value); }
std::uint64_t loratoh(std::uint64_t _value) { return le64toh(_value); }

#define MHDR_SIZE sizeof(std::uint8_t)
#define MIC_SIZE sizeof(std::uint32_t)
#define PHYPAYLOAD_SIZE(x) (MHDR_SIZE + (x) + MIC_SIZE)

#define DEVADDR_SIZE sizeof(std::uint32_t)
#define FCTRL_SIZE sizeof(std::uint8_t)
#define FCNT_SIZE sizeof(std::uint16_t)
#define FHDR_MIN_SIZE (DEVADDR_SIZE + FCTRL_SIZE + FCNT_SIZE)
#define MACPAYLOAD_MIN_SIZE FHDR_MIN_SIZE

#define JOINEUI_SIZE sizeof(std::uint64_t)
#define DEVEUI_SIZE sizeof(std::uint64_t)
#define DEVNONCE_SIZE sizeof(std::uint16_t)
#define JOIN_REQUEST_SIZE (JOINEUI_SIZE + DEVEUI_SIZE + DEVNONCE_SIZE)

#define RJTYPE_SIZE sizeof(std::uint8_t)
#define NETID_SIZE (sizeof(std::uint8_t) + sizeof(std::uint16_t))
#define RJCOUNT_SIZE sizeof(std::uint16_t)
#define REJOIN_REQUEST_02_SIZE                                                 \
	(RJTYPE_SIZE + NETID_SIZE + DEVEUI_SIZE + RJCOUNT_SIZE)
#define REJOIN_REQUEST_1_SIZE                                                  \
	(RJTYPE_SIZE + JOINEUI_SIZE + DEVEUI_SIZE + RJCOUNT_SIZE)

#define PHYPAYLOAD_MIN_SIZE PHYPAYLOAD_SIZE(MACPAYLOAD_MIN_SIZE)

struct __attribute__((packed)) FrameHeader
{
	std::uint32_t deviceAddress;
	std::uint8_t frameControl;
	std::uint16_t frameCounter;
};

struct __attribute__((packed)) JoinRequest
{
	std::uint64_t joinEUI;
	std::uint64_t deviceEUI;
	std::uint16_t deviceNonce;
};

struct __attribute__((packed)) RejoinRequestType02
{
	std::uint16_t netIDLSB;
	std::uint8_t netIDMSB;
	std::uint64_t deviceEUI;
	std::uint16_t rejoinCount;
};

struct __attribute__((packed)) RejoinRequestType1
{
	std::uint64_t joinEUI;
	std::uint64_t deviceEUI;
	std::uint16_t rejoinCount;
};

struct __attribute__((packed)) RejoinRequest
{
	std::uint8_t rejoinType;

	union {
		RejoinRequestType02 type02;
		RejoinRequestType1 type1;
	};
};

enum LoRaWAN::Direction getDirection(enum LoRaWAN::MessageType _messageType)
{
	switch (_messageType) {
	case LoRaWAN::TYPE_JOIN_REQUEST:
	// fallthrough
	case LoRaWAN::TYPE_UNCONFIRMED_DATA_UP:
	// fallthrough
	case LoRaWAN::TYPE_CONFIRMED_DATA_UP:
	// fallthrough21344,
	case LoRaWAN::TYPE_REJOIN_REQUEST:
		return LoRaWAN::DIR_UPLINK;

	case LoRaWAN::TYPE_JOIN_ACCEPT:
	// fallthrough
	case LoRaWAN::TYPE_UNCONFIRMED_DATA_DOWN:
	// fallthrough
	case LoRaWAN::TYPE_CONFIRMED_DATA_DOWN:
		return LoRaWAN::DIR_DOWNLINK;

	default:
		return LoRaWAN::DIR_UNKNOWN;
	}
}
}

bool LoRaWAN::isData(void) const
{
	switch (messageType) {
	case TYPE_UNCONFIRMED_DATA_DOWN:
	// fallthrough
	case TYPE_UNCONFIRMED_DATA_UP:
	// fallthrough
	case TYPE_CONFIRMED_DATA_DOWN:
	// fallthrough
	case TYPE_CONFIRMED_DATA_UP:
	// fallthrough
	case TYPE_PROPRIETARY:
		return true;

	default:
		return false;
	}
}

LoRaWAN::LoRaWAN(const std::string &_payload)
	: LoRaWAN(reinterpret_cast<const std::uint8_t *>(_payload.data()),
		  _payload.size()){};

LoRaWAN::LoRaWAN(const std::uint8_t *_payload, std::size_t _size)
	: payload(_payload), size(_size), messageType(TYPE_UNKNOWN),
	  direction(DIR_UNKNOWN), majorVersion(0), deviceAddress(0),
	  frameControl(0), hasAdaptativeDataRate(false),
	  hasADRAcknowlegdementRequest(false), hasAcknowlegdement(false),
	  hasFramePending(false), hasClassB(false), frameOptionsLength(0),
	  frameCounter(0), hasFramePort(false), framePort(0), joinEUI(0),
	  deviceEUI(0), deviceNonce(0), rejoinType(0), netID(0), rejoinCount(0)
{
	if (size < PHYPAYLOAD_MIN_SIZE) {
		logr::warning << "Invalid PHY payload size" << logr::endl;
		return;
	}

	messageType = static_cast<enum MessageType>((payload[0] >> 5) & 0x07);
	direction = getDirection(messageType);
	majorVersion = (payload[0] >> 0) & 0x03;

	if (majorVersion > MAJOR_VERSION) {
		logr::warning << "Invalid major version" << logr::endl;
		return;
	}

	if (isData()) {
		const FrameHeader *frameHeader =
			reinterpret_cast<const FrameHeader *>(payload +
							      MHDR_SIZE);

		deviceAddress = loratoh(frameHeader->deviceAddress);
		frameControl = frameHeader->frameControl;
		frameCounter = loratoh(frameHeader->frameCounter);

		if (frameControl & 1 << 7) {
			hasAdaptativeDataRate = true;
		}

		if (direction == DIR_UPLINK && frameControl & 1 << 6) {
			hasADRAcknowlegdementRequest = true;
		}

		if (frameControl & 1 << 5) {
			hasAcknowlegdement = true;
		}

		if (direction == DIR_DOWNLINK && frameControl & 1 << 4) {
			hasFramePending = true;
		}

		if (direction == DIR_UPLINK && frameControl & 1 << 4) {
			hasClassB = true;
		}

		frameOptionsLength = frameControl & 0x0F;

		if (size < PHYPAYLOAD_MIN_SIZE + frameOptionsLength) {
			logr::warning << "Invalid frame options length"
				      << logr::endl;
			return;

		} else if (size > PHYPAYLOAD_MIN_SIZE + frameOptionsLength) {
			hasFramePort = true;
			framePort = payload[MHDR_SIZE + MACPAYLOAD_MIN_SIZE +
					    frameOptionsLength];
		}

	} else if (messageType == TYPE_JOIN_REQUEST) {
		if (size != PHYPAYLOAD_SIZE(JOIN_REQUEST_SIZE)) {
			logr::warning << "Invalid join request size"
				      << logr::endl;
			return;
		}

		const JoinRequest *joinRequest =
			reinterpret_cast<const JoinRequest *>(payload +
							      MHDR_SIZE);

		joinEUI = loratoh(joinRequest->joinEUI);
		deviceEUI = loratoh(joinRequest->deviceEUI);
		deviceNonce = loratoh(joinRequest->deviceNonce);

	} else if (messageType == TYPE_REJOIN_REQUEST) {
		const RejoinRequest *rejoinRequest =
			reinterpret_cast<const RejoinRequest *>(payload +
								MHDR_SIZE);

		rejoinType = rejoinRequest->rejoinType;

		switch (rejoinType) {
		case 0:
		// fallthrough
		case 2:
			if (size != PHYPAYLOAD_SIZE(REJOIN_REQUEST_02_SIZE)) {
				logr::warning << "Invalid rejoin request size"
					      << logr::endl;
				return;
			}

			netID = rejoinRequest->type02.netIDMSB << 16 |
				loratoh(rejoinRequest->type02.netIDLSB);
			deviceEUI = loratoh(rejoinRequest->type02.deviceEUI);
			rejoinCount =
				loratoh(rejoinRequest->type02.rejoinCount);
			break;

		case 1:
			if (size != PHYPAYLOAD_SIZE(REJOIN_REQUEST_1_SIZE)) {
				logr::warning << "Invalid rejoin request size"
					      << logr::endl;
				return;
			}

			joinEUI = loratoh(rejoinRequest->type1.joinEUI);
			deviceEUI = loratoh(rejoinRequest->type1.deviceEUI);
			rejoinCount = loratoh(rejoinRequest->type1.rejoinCount);
			break;

		default:
			logr::warning << "Invalid rejoin type" << logr::endl;
			return;
		}
	}
}

namespace {
std::string to_string(enum LoRaWAN::MessageType _messageType)
{
	switch (_messageType) {
	case LoRaWAN::TYPE_JOIN_REQUEST:
		return "Join Request";
	case LoRaWAN::TYPE_JOIN_ACCEPT:
		return "Join Accept";
	case LoRaWAN::TYPE_UNCONFIRMED_DATA_UP:
		return "Unconfirmed Data Up";
	case LoRaWAN::TYPE_UNCONFIRMED_DATA_DOWN:
		return "Unconfirmed Data Down";
	case LoRaWAN::TYPE_CONFIRMED_DATA_UP:
		return "Confirmed Data Up";
	case LoRaWAN::TYPE_CONFIRMED_DATA_DOWN:
		return "Confirmed Data Down";
	case LoRaWAN::TYPE_REJOIN_REQUEST:
		return "Rejoin Request";
	case LoRaWAN::TYPE_PROPRIETARY:
		return "Proprietary";
	default:
		return "Unknown";
	}
}
}

std::ostream &operator<<(std::ostream &_os, const LoRaWAN &_lorawan)
{
	_os << to_string(_lorawan.messageType);

	if (_lorawan.majorVersion > LoRaWAN::MAJOR_VERSION) {
		_os << ", Major " << static_cast<int>(_lorawan.majorVersion);
		return _os;
	}

	if (_lorawan.isData()) {
		_os << ", DevAddr " << to_hex(_lorawan.deviceAddress);

		if (_lorawan.frameControl) {
			_os << ", FCtrl";

			if (_lorawan.hasAdaptativeDataRate) {
				_os << " [ADR]";
			}

			if (_lorawan.hasADRAcknowlegdementRequest) {
				_os << " [ADRACKReq]";
			}

			if (_lorawan.hasAcknowlegdement) {
				_os << " [ACK]";
			}

			if (_lorawan.hasFramePending) {
				_os << " [FPending]";
			}

			if (_lorawan.hasClassB) {
				_os << " [ClassB]";
			}
		}

		_os << ", FCnt " << _lorawan.frameCounter;

		if (_lorawan.hasFramePort) {
			_os << ", FPort "
			    << static_cast<int>(_lorawan.framePort);
		}

	} else if (_lorawan.messageType == LoRaWAN::TYPE_JOIN_REQUEST) {
		if (_lorawan.size == PHYPAYLOAD_SIZE(JOIN_REQUEST_SIZE)) {
			_os << ", JoinEUI " << to_hex(_lorawan.joinEUI);
			_os << ", DevEUI " << to_hex(_lorawan.deviceEUI);
			_os << ", DevNonce " << _lorawan.deviceNonce;
		}

	} else if (_lorawan.messageType == LoRaWAN::TYPE_REJOIN_REQUEST) {
		if (_lorawan.size == PHYPAYLOAD_SIZE(REJOIN_REQUEST_02_SIZE) &&
		    (_lorawan.rejoinType == 0 || _lorawan.rejoinType == 2)) {
			_os << ", RejoinType "
			    << static_cast<int>(_lorawan.rejoinType);
			_os << ", NetID "
			    << to_hex(_lorawan.netID, NETID_SIZE << 1);
			_os << ", DevEUI " << to_hex(_lorawan.deviceEUI);
			_os << ", RJcount0 " << _lorawan.rejoinCount;

		} else if (_lorawan.size ==
				   PHYPAYLOAD_SIZE(REJOIN_REQUEST_1_SIZE) &&
			   _lorawan.rejoinType == 1) {
			_os << ", RejoinType "
			    << static_cast<int>(_lorawan.rejoinType);
			_os << ", JoinEUI " << to_hex(_lorawan.joinEUI);
			_os << ", DevEUI " << to_hex(_lorawan.deviceEUI);
			_os << ", RJcount1 " << _lorawan.rejoinCount;
		}
	}

	return _os;
}
}
