// Copyright (c) 2018 Kerlink SA. All rights reserved.

#ifndef _UTILS_SERVICE_H_
#define _UTILS_SERVICE_H_

#include <chrono>
#include <condition_variable>
#include <functional>
#include <mutex>
#include <stdexcept>
#include <thread>
#include <vector>

#include "utils/logger.h"

namespace service {
class Timer
{
public:
	Timer(void) : m_running(false) {}
	~Timer(void) { shutdown(); }

	// Not copyable
	Timer(const Timer &) = delete;
	Timer &operator=(const Timer &) = delete;

	// Not movable
	Timer(Timer &&) = delete;
	Timer &operator=(Timer &&) = delete;

	template <class Rep, class Period>
	bool sleep_for(const std::chrono::duration<Rep, Period> &_duration)
	{
		std::unique_lock<std::mutex> lock(m_mutex);
		m_running = true;
		return !m_condition.wait_for(
			lock, _duration, [&] { return !m_running; });
	}

	void shutdown(void)
	{
		{
			std::lock_guard<std::mutex> lock(m_mutex);
			m_running = false;
		}

		m_condition.notify_all();
	}

private:
	std::condition_variable m_condition;
	std::mutex m_mutex;
	bool m_running;
};

class Service
{
private:
	struct Worker
	{
		Worker(void) {}
		~Worker(void)
		{
			m_timer.shutdown();
			join();
		}

		// Not copyable
		Worker(const Worker &) = delete;
		Worker &operator=(const Worker &) = delete;

		// Not movable
		Worker(Worker &&) = delete;
		Worker &operator=(Worker &&) = delete;

		bool run(std::function<void(void)> _handler)
		{
			try {
				m_thread = std::thread(_handler);
				return true;

			} catch (const std::exception &e) {
				logr::error
					<< "Failed to run worker: " << e.what()
					<< logr::endl;
				return false;
			}
		}

		bool join(void)
		{
			try {
				if (m_thread.joinable()) {
					m_thread.join();
				}

				return true;

			} catch (const std::exception &e) {
				logr::error
					<< "Failed to join worker: " << e.what()
					<< logr::endl;
				return false;
			}
		}

		std::thread m_thread;
		Timer m_timer;
	};

	std::vector<Worker *> m_workers;

	Worker *this_worker(void)
	{
		for (auto *worker : m_workers) {
			if (worker->m_thread.get_id() ==
			    std::this_thread::get_id()) {
				return worker;
			}
		}

		throw std::runtime_error("Invalid worker");
	}

protected:
	template <class Rep, class Period>
	bool sleep_for(const std::chrono::duration<Rep, Period> &_duration)
	{
		try {
			return this_worker()->m_timer.sleep_for(_duration);

		} catch (const std::exception &e) {
			std::this_thread::sleep_for(_duration);
			return true;
		}
	}

	bool m_running;

public:
	Service(void) : m_running(false) {}

	// Not copyable
	Service(const Service &) = delete;
	Service &operator=(const Service &) = delete;

	// Not movable
	Service(Service &&) = delete;
	Service &operator=(Service &&) = delete;

	virtual ~Service(void)
	{
		shutdown();
		join();
		for (auto *worker : m_workers) {
			delete worker;
		}
	}

	virtual bool start(void) = 0;

	virtual bool running(void) const { return m_running; }

	virtual void shutdown(void)
	{
		m_running = false;

		for (auto *worker : m_workers) {
			worker->m_timer.shutdown();
		}
	}

	bool join(void)
	{
		bool status = true;

		for (auto *worker : m_workers) {
			if (!worker->join()) {
				status = false;
			}
		}

		return status;
	}

protected:
	bool add_worker(std::function<bool(void)> _handler)
	{
		m_running = true;

		try {
			m_workers.emplace_back(new Worker());

			std::function<void(void)> handler = [this, _handler] {
				while (running() && _handler())
					;
			};

			if (!m_workers.back()->run(handler)) {
				m_workers.pop_back();
				return false;
			}

			return true;

		} catch (const std::exception &e) {
			logr::error << "Failed to add worker: " << e.what()
				    << logr::endl;
			return false;
		}
	}

	template <typename Callable>
	bool add_workers(Callable _callable)
	{
		return add_worker(_callable);
	}

	template <typename Callable, typename... Args>
	bool add_workers(Callable _callable, Args... _args)
	{
		if (!add_workers(_callable)) {
			return false;
		}

		return add_workers(_args...);
	}
};
}

#endif
